#!/usr/bin/env python
import os 
import sys
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')

import wuml
import torch
import torch.nn as nn

#	Set this variable in the very beginning to cpu to ignore the gpu 
pytorch_device = None

#	Pytorch helpers
#	----------------------------------------------
def get_current_device():
	if wuml.pytorch_device is not None: return wuml.pytorch_device

	if torch.cuda.is_available(): wuml.pytorch_device = 'cuda'
	else: wuml.pytorch_device = 'cpu'
	return wuml.pytorch_device

#	Loss function
#	----------------------------------------------
def CrossEntropyLoss(y, ŷ):
	lossFun = nn.CrossEntropyLoss() 

	#weird from pytorch, dim of y is 20, and ŷ is 20x3
	# it seems that the loss function performs
	#	1. softmax
	return lossFun(ŷ, y) 	

def MSELoss(y, ŷ):
	mseFun = nn.MSELoss()
	if ŷ.shape != y.shape:
		raise ValueError('Error: \n\tIn wuml.MSELoss y.shape=%s and ŷ.shape=%s have dimension mismatch [maybe ŷ.squeeze() is needed].'%(str(ŷ.shape), str(y.shape)))

	return mseFun(ŷ, y)

def softmax(x, turn_into_label=False):
	m = nn.Softmax()
	xout = m(x)

	if turn_into_label:
		_, yout = torch.max(xout, 1)
		return yout
	else:
		return xout





from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
from wuml.type_check import *
import numpy as np


def impute(data, imputerType='iterative', ignore_first_index_column=False):
	X = ensure_numpy(data)

	if ignore_first_index_column:
		c1 = np.atleast_2d(X[:,0]).T
		X = X[:,1:]

	imp_mean = IterativeImputer(random_state=0)
	imp_mean.fit(X)
	Xi = imp_mean.transform(X)

	if ignore_first_index_column:
		Xi = np.hstack((c1,Xi))

	Xout = ensure_data_type(Xi, type_name=type(data).__name__)	# returns the same data type as input type
	if wtype(Xout) == 'wData' and wtype(data) == 'wData':
		Xout.rename_columns(data.get_column_names_as_a_list())
		Xout.label_column_name = data.label_column_name

		Xout.Y = data.Y

	return Xout
	




#!/usr/bin/env python

import os 
import sys

if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')

import types
import torch
import pickle
import numpy as np
import wuml

import wplotlib
import marshal, types
import pandas as pd
from pathlib import Path
from sklearn.metrics import accuracy_score
from IPython.display import clear_output, HTML, Math
from wuml.type_check import *
from IPython.display import display_html
from itertools import chain,cycle
  


#	display 2 dataframes side by side
def DF_display_side_by_side(*args,titles=cycle([''])):
	#	Example usage
	#df1 = pd.DataFrame(np.arange(12).reshape((3,4)),columns=['A','B','C','D',])
	#df2 = pd.DataFrame(np.arange(16).reshape((4,4)),columns=['A','B','C','D',])
	#display_side_by_side(df1,df2,df1, titles=['Foo','Foo Bar']) 

	html_str=''
	for df,title in zip(args, chain(titles,cycle(['</br>'])) ):
		html_str+='<th style="text-align:center"><td style="vertical-align:top">'
		html_str+=f'<h2 style="text-align: center;">{title}</h2>'
		html_str+=df.to_html().replace('table','table style="display:inline"')
		html_str+='</td></th>'
	display_html(html_str,raw=True)


def print_two_matrices_side_by_side(M1, M2, title1=None, title2=None, auto_print=True, rounding=3):

	if wuml.isnotebook():			# and wtype(M1) =='DataFrame' and wtype(M2) =='DataFrame':
		M1 = ensure_DataFrame(M1)
		M2 = ensure_DataFrame(M2)
		wuml.DF_display_side_by_side(M1, M2,titles=[title1, title2])
	else:
		eK = wuml.pretty_np_array(M1, front_tab='', title=title1, auto_print=False)
		eQ = wuml.pretty_np_array(M2, front_tab='\t', title=title2, auto_print=False)
		wuml.block_two_string_concatenate(eK, eQ, spacing='\t\t', add_titles=[], auto_print=auto_print)



def jupyter_print(value, display_all_rows=False, display_all_columns=False, font_size=1, latex=False, endString=None, testing_mode=''):
	cmds = wuml.get_commandLine_input()
	if testing_mode == 'disabled' or cmds[1] == 'disabled': return

	#font_size is from 1 to 6
	font_size = int(6 - font_size)
	if font_size > 6: font_size = 6
	if font_size < 0: font_size = 1

	if wtype(value) == 'result_table': 
		value = value.df

	if display_all_columns: pd.set_option('display.max_columns', None) 
	if display_all_rows: pd.set_option('display.max_rows', None)
	if wuml.isnotebook():
		if wtype(value) == 'DataFrame': 
			display(value)
		elif wtype(value) == 'ndarray': 
			display(ensure_DataFrame(value))
		elif wtype(value) == 'wData': 
			display(value.df)
		elif wtype(value) == 'Index': 
			value = str(value.tolist())
			str2html = '<html><body><h%d>%s</h%d></body></html>'%(font_size, value, font_size)
			display(HTML(data=str2html))
		elif wtype(value) == 'tuple': 
			value = str(value)
			str2html = '<html><body><h%d>%s</h%d></body></html>'%(font_size, value, font_size)
			display(HTML(data=str2html))
		elif wtype(value) == 'str': 
			#value = value.replace('\r','<br>')
			#value = value.replace('\n','<br>')
			#value = value.replace('\t','&nbsp &nbsp')
			if latex:
				display(Math(r'%s'%value))
			else:
				print(value)
				#wuml.write_to_current_line(value)
				#str2html = '<html><body><h%d>%s</h%d></body></html>'%(font_size, value, font_size)
				#display(HTML(data=str2html))
		else:
			if wtype(value) == 'Tensor': value = wuml.ensure_numpy(value)
			print(value)

		pd.set_option('display.max_rows', 10)
	else:
		if wtype(value) == 'Tensor': value = wuml.ensure_numpy(value)
		print(value)
	if endString is not None: print(endString)

def set_terminal_print_options(precision=3):
	np.set_printoptions(precision=precision)
	np.set_printoptions(linewidth=400)
	np.set_printoptions(suppress=True)

	torch.set_printoptions(edgeitems=3)
	torch.set_printoptions(sci_mode=False)
	torch.set_printoptions(precision=precision)
	torch.set_printoptions(linewidth=400)

	pd.set_option('display.width', 1000)
	pd.set_option('display.max_columns', 80)
	#pd.set_option("display.max_rows", 500)
	pd.set_option("display.precision", precision)

def csv_load(xpath, ypath=None, shuffle_samples=False):
	X = np.genfromtxt(	xpath, delimiter=',')
	Y = None
	if ypath is not None: 
		Y = np.genfromtxt(ypath, delimiter=',')

	if shuffle_samples:
		if Y is None:
			X = shuffle(X, random_state=0)
		else:
			X, Y = shuffle(X, Y, random_state=0)

	if Y is None: return X
	else: return [X,Y]


def csv_out(X, path, add_row_indices=False, include_column_names=False, float_format='%.4f'):
	X = wuml.ensure_wData(X)
	X.to_csv(path, add_row_indices=add_row_indices, include_column_names=include_column_names, float_format=float_format)


def label_csv_out(y, path, float_format='%.4f'):
	np.savetxt(path, y, delimiter=',', fmt=float_format) 


def read_txt_file(path):
	f = open(path, 'r')
	content = f.read()
	f.close()
	return content

def save_torch_network(network_obj, path):
	N = network_obj
	netData  = N.output_network_data_for_storage()
	pickle_dump(netData, path)

def load_torch_network(path, load_as_cpu_or_gpu=None): # load_as_cpu_or_gpu, if set to 'cpu' or 'cuda', it will try to force it 
	if load_as_cpu_or_gpu is not None:
		wuml.pytorch_device = load_as_cpu_or_gpu

	netData = pickle_load(path)
	
	if netData['name'] == 'basicNetwork':
		net = wuml.basicNetwork(None, None, pickled_network_info=netData)
	elif netData['name'] == 'combinedNetwork':
		code = marshal.loads(netData['network_behavior_on_call'])
		network_behavior_on_call = types.FunctionType(code, globals(), "network_behavior_on_call")

		code = marshal.loads(netData['costFunction'])
		costFunction = types.FunctionType(code, globals(), "costFunction")
		net = wuml.combinedNetwork(None, None, None, costFunction, network_behavior_on_call, pickled_network_info=netData, data_mean=netData['μ'], data_std=netData['σ'])

		if 'explainer' in netData.keys():
			stored_reference_data = wuml.ensure_wData(netData['reference_data'] , column_names=netData['column_names'])
			net.explainer = wuml.explainer(stored_reference_data, net, explainer_algorithm=netData['explainer_algorithm'], which_model_output_to_use=netData['which_model_output_to_use'])


	elif netData['name'] == 'autoencoder':
		code = marshal.loads(netData['costFunction'])
		costFunction = types.FunctionType(code, globals(), "costFunction")
		net = wuml.autoencoder(None, None, costFunction=costFunction, pickled_network_info=netData)
	else:
		raise ValueError('\n\tError loading torch network: %s is unrecognized network class'%netData['name'])

	return net

#
#func(10)  # gives 100



def pickle_dump(obj, path):
	pickle.dump( obj, open( path, "wb" ) )
   
def pickle_load(pth):
	return pickle.load( open( pth, "rb" ) )

def copy_file(src, destination):
	copyfile(src, destination)

def write_to(txt, path):
	f = open(path, 'w')
	f.write(txt)
	f.close()

def go_up_1_directory(full_path):	# path should be a string
	return str(Path(full_path).parents[0])

def initialize_empty_folder(path):
	ensure_path_exists(path)
	remove_files(path)

def path_exists(path):
	if os.path.exists(path): return True
	return False

def list_all_files_in_directory(path, with_extension=None):
	all_files = os.listdir(path)

	if with_extension is not None:
		new_all_files = []
		for f in all_files:
			file_extension = Path(f).suffix
			if with_extension == file_extension:
				new_all_files.append(f)

		all_files = new_all_files

	return all_files

def ensure_path_exists(path):
	if os.path.exists(path): return True
	os.mkdir(path)
	return False

def path_list_exists(path_list):
	for i in path_list:
		if os.path.exists(i) == False: 
			return False

	return True

# path_prefix can be a string or a list of string
# if list, then it will check the existence given the prefix
def append_prefix_to_path(path_prefix, fname):
	if wtype(path_prefix) == 'str':
		ypth = path_prefix + fname
	elif wtype(path_prefix) == 'list':
		for pre in path_prefix:
			ypth = pre + fname
			if wuml.path_exists(ypth):
				break
	else:
		raise ValueError('path_prefix must be a string or a list of strings')

	return ypth


def create_file(path):
	fin = open(path,'w')
	fin.close()

def delete_file(path):
	if os.path.exists(path):
		os.remove(path)

def remove_files(folder_path):	# make sure to end the path with /
	if not os.path.exists(folder_path): return 

	file_in_tmp = os.listdir(folder_path)
	for i in file_in_tmp:
		if os.path.isfile(folder_path + i):
			os.remove(folder_path + i)





#	Terminal Printing

def print_status(percent):
	if get_commandLine_input()[1] == 'disabled': return
	clear_current_line()
	write_to_current_line('Completion Status: %.3f'%percent)
	sys.stdout.flush()

def clear_current_line():
	if get_commandLine_input()[1] == 'disabled': return
	#print(..., end="")
	#sys.stdout.write('\033[2K\033[1G')
	sys.stdout.write("\r")		#jump to the beginning of line
	sys.stdout.write("\033[K")	#erase to the end of line
	#sys.stdout.write("\033[F")		move cursor to beginning of previous line
	#“\033[A” – move cursor up one line.
	
	sys.stdout.flush()

def write_to_current_line(txt):
	if get_commandLine_input()[1] == 'disabled': return

	clear_current_line()
	clear_output(wait = True)
	sys.stdout.write(txt)
	sys.stdout.flush()

def clear_previous_line(num_of_lines=None):
	if get_commandLine_input()[1] == 'disabled': return
	if num_of_lines is None: num_of_lines = 1
	for i in range(num_of_lines):
		clear_current_line()
		sys.stdout.write("\r")
		sys.stdout.write("\033[F")
		sys.stdout.flush()
		clear_current_line()

def loss_optimization_printout(db, epoch, avgLoss, avgGrad, epoc_loop, slope):
	sys.stdout.write("\r\t\t%d/%d, MaxLoss : %f, AvgGra : %f, progress slope : %f" % (epoch, epoc_loop, avgLoss, avgGrad, slope))
	sys.stdout.flush()

def dictionary_to_str(dic):
	outstr = ''
	for i,j in dic.items():
		if type(j) == str: outstr += 	('\t\t' + i + ' : ' + str(j) + '\n')
		elif type(j) == np.float64: outstr += 	('\t\t' + i + ' : ' + str(j) + '\n')
		elif type(j) == bool: outstr += ('\t\t' + i + ' : ' + str(j) + '\n')
		elif type(j) == type: outstr += ('\t\t' + i + ' : ' + j.__name__ + '\n')
		elif type(j) == types.FunctionType: outstr += ('\t\t' + i + ' : ' + j.__name__ + '\n')
		elif type(j) == float: outstr += ('\t\t' + i + ' : ' + str(j) + '\n')
		elif type(j) == int: outstr += 	('\t\t' + i + ' : ' + str(j) + '\n')
		else:
			print('%s , %s is not recognized'%(i, str(type(j))))
			import pdb; pdb.set_trace()	

	return outstr

def output_two_columns_side_by_side(col_1, col_2, labels=None, rounding=3):
	col_1 = np.atleast_2d(wuml.ensure_numpy(col_1, rounding=rounding))
	col_2 = np.atleast_2d(wuml.ensure_numpy(col_2, rounding=rounding))

	if col_1.shape[0] == 1: col_1 = col_1.T
	if col_2.shape[0] == 1: col_2 = col_2.T

	if wuml.isnotebook():
		output = wuml.ensure_wData(np.hstack((col_1, col_2)), column_names=labels)
		jupyter_print(output, display_all_rows=True)
	else:
		output = wuml.pretty_np_array(np.hstack((col_1, col_2)))

		if labels is not None:
			output = wuml.pretty_np_array(labels) + output
		wuml.jupyter_print(output)

		return output


def get_commandLine_input(num_of_args=10):
	command_list = []	
	for i in range(num_of_args):
		try:
			command_list.append(sys.argv[i])
		except:
			command_list.append('')

	return command_list

def output_regression_result(y, ŷ, write_path=None, sort_by='none', ascending=False, print_out=['histograms', 'mean absolute error', 'true label vs prediction table']):
	y = np.atleast_2d(wuml.ensure_numpy(y, rounding=2))
	ŷ = np.atleast_2d(wuml.ensure_numpy(ŷ, rounding=2))

	if y.shape[0] == 1: y = y.T
	if ŷ.shape[0] == 1: ŷ = ŷ.T
	
	# Draw Histogram
	Δy = np.absolute(ŷ - y)

	if 'mean absolute error' in print_out:
		avg_Δ = 'Avg error: %.4f'%(np.sum(Δy)/Δy.shape[0])
		jupyter_print(avg_Δ)

	if 'histograms' in print_out:
		wplotlib.histograms(Δy, num_bins=15, title='Histogram of Errors', xlabel='Error Amount', ylabel='Error count',
					facecolor='blue', α=0.5, path=None)

	df = pd.DataFrame(np.hstack((y, ŷ, Δy)), columns=['y', 'ŷ', 'Δy'])
	if sort_by == 'error':
		df = df.sort_values('Δy', ascending=ascending)
	elif sort_by == 'label':
		df = df.sort_values('y', ascending=ascending)
	elif sort_by == 'output':
		df = df.sort_values('ŷ', ascending=ascending)

	if not wuml.isnotebook():
		if 'true label vs prediction table' in print_out:
			jupyter_print(df, display_all_rows=True, display_all_columns=True)
	
	if write_path is not None: wuml.write_to(str(df), write_path)
	return wuml.ensure_wData(df)


class summarize_regression_result:
	def __init__(self, y, ŷ, print_out=['avg absolute error', 'true v predict'], rounding=3):
		y = np.atleast_2d(wuml.ensure_numpy(y, rounding=rounding))
		ŷ = np.atleast_2d(wuml.ensure_numpy(ŷ, rounding=rounding))
	
		if y.shape[0] == 1: y = y.T
		if ŷ.shape[0] == 1: ŷ = ŷ.T
		
		self.y = y
		self.ŷ = ŷ
		self.Δy = np.round(np.absolute(self.ŷ - self.y), 3)
		self.mean_absolute_error = self.avg_error()

		if print_out is not None:
			if 'avg absolute error' in print_out:
				jupyter_print('Mean Absolute Error: %.4f'%self.mean_absolute_error)

			if 'true v predict' in print_out:
				self.true_vs_predict(print_out=True)

		self.check_for_errors(print_out)

	def check_for_errors(self, print_out):
		if print_out is None: return
		for i in print_out:
			if i not in ['avg absolute error', 'true v predict']:
				raise ValueError('\n\tError in summarize_regression_result: %s is not recognized print_out option'%i)


	def avg_error(self):
		Δy = self.Δy
		avg_Δ = np.sum(Δy)/Δy.shape[0]
		return avg_Δ

	def error_histogram(self):
		Δy = self.Δy
		avg_Δ = 'Avg error: %.4f\n\n'%(np.sum(Δy)/Δy.shape[0])
	
		H = wplotlib.histograms(Δy, num_bins=15, title='Histogram of Errors', 
					xlabel='Error Amount', ylabel='Error count',
					facecolor='blue', α=0.5, path=None)
		
	def true_vs_predict(self, write_path=None, sort_by='none', ascending=False, print_out=False):
		cnames = np.array(['y', 'ŷ', 'Δy'])
		Yjoin = np.hstack((self.y, self.ŷ, self.Δy))
		df = pd.DataFrame(Yjoin, columns=cnames)

		# round the data frame decimals correctly
		y_ndeci = len(str(np.squeeze(self.y)[0]).split('.')[1])
		ŷ_ndeci = len(str(np.squeeze(self.ŷ)[0]).split('.')[1])
		Δy_ndeci = len(str(np.squeeze(self.Δy)[0]).split('.')[1])
		df = df.round({'y':y_ndeci, 'ŷ':ŷ_ndeci, 'Δy':Δy_ndeci})

		if sort_by == 'error':
			df = df.sort_values('Δy', ascending=ascending)
		elif sort_by == 'label':
			df = df.sort_values('y', ascending=ascending)
		elif sort_by == 'output':
			df = df.sort_values('ŷ', ascending=ascending)

		if print_out:
			jupyter_print(df, display_all_rows=True, display_all_columns=True)

		return wuml.ensure_wData(df)


class summarize_classification_result:
	def __init__(self, y, ŷ, print_out=['accuracy', 'true v predict labels']):

		y = np.atleast_2d(wuml.ensure_numpy(y, rounding=2))
		ŷ = np.atleast_2d(wuml.ensure_numpy(ŷ, rounding=2))
	
		if y.shape[0] == 1: y = y.T
		if ŷ.shape[0] == 1: ŷ = ŷ.T

		self.y = y
		self.ŷ = ŷ
		self.side_by_side_Y = np.hstack((self.y, self.ŷ))
		self.Δy = np.absolute(self.ŷ - self.y)

		self.accuracy = self.get_accuracy()
		if len(np.unique(self.y)) == 2: 
			self.Precision = wuml.precision(self.y, self.ŷ)
			self.Recall = wuml.recall(self.y, self.ŷ)

		if wuml.get_commandLine_input()[1] == 'disabled': return
		# Printing out the result
		if print_out is not None:
			if 'accuracy' in print_out:
				jupyter_print('Classification Accuracy: %.4f'%self.accuracy)
	
			if is_binary_label(y):
				jupyter_print('Precision: %.4f (probability that a positive prediction is correct)'%self.Precision)
				jupyter_print('Recall: %.4f (probability that we catch a positive event)'%self.Recall)

			if 'true v predict labels' in print_out:
				self.true_vs_predict(print_out=True)

		self.check_for_errors(print_out)

	def check_for_errors(self, print_out):
		if print_out is None: return
		for i in print_out:
			if i not in ['accuracy', 'true v predict labels']:
				raise ValueError('Error in summarize_classification_result: %s is not recognized print_out option'%i)

	def get_accuracy(self):
		Acc= accuracy_score(self.y, self.ŷ)
		return Acc
		
	def true_vs_predict(self, add_to_label='', write_path=None, sort_by='none', ascending=False, print_out=False):
		cnames = np.array(['y'+add_to_label, 'ŷ'+ add_to_label, 'Δy'+add_to_label])
		Yjoin = np.hstack((self.y, self.ŷ, self.Δy))
		df = pd.DataFrame(Yjoin, columns=cnames)

		if sort_by == 'error':
			df = df.sort_values('Δy', ascending=ascending)
		elif sort_by == 'label':
			df = df.sort_values('y', ascending=ascending)
		elif sort_by == 'output':
			df = df.sort_values('ŷ', ascending=ascending)

		if print_out:
			jupyter_print(df, display_all_rows=True, display_all_columns=True)

		return wuml.ensure_wData(df)


	

class result_table:
	def __init__(self, column_names, data=None):
		self.column_names = column_names
		if data is None:
			self.df = pd.DataFrame(columns=column_names)
		else:
			self.df = pd.DataFrame(data, columns=column_names)


	def add_row(self, row_data):
		zP = zip(self.column_names, row_data)
		D = {}
		for i,j in zP:
			D[i] = j

		self.df = self.df.append(D, ignore_index = True)

	def __str__(self): 
		return str(self.df)


	def get_column(self, column):
		if type(column).__name__ == 'int': 
			return wuml.ensure_wData(self.df.iloc[:,column])
		
		return wuml.ensure_wData(self.df[column])




#!/usr/bin/env python

import os
import sys
if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')

from wplotlib import histograms
import numpy as np
import wuml
import torch
from wuml.type_check import *


#str.ljust(s, width[, fillchar])
#str.rjust(s, width[, fillchar])
#str.center(s, width[, fillchar])


def pretty_np_array(m, front_tab='', verticalize=False, title=None, auto_print=False, end_space='', round_value=3):

	M = ensure_wData(m)
	M.round(rounding=round_value)

	#if wtype(m) == 'float64': 
	#	import pdb; pdb.set_trace()

	if auto_print: wuml.jupyter_print(M)
	else: 
		out_str = str(M)
		L1 = out_str.split('\n')
		L1_max_width = len(max(L1, key=len))
		if wtype(title) == 'str':
			t1 = str.center(title, L1_max_width)
			out_str = t1 + '\n' + out_str + end_space
		else:
			out_str = out_str + end_space

		return out_str


#	This portion is deprecated, we now use dataframe instead of np array.

#	cName = None
#	if wtype(m) =='DataFrame': 
#		out_str = str(m.round(round_value))
#	else:
#		m = ensure_wData(m)
#		cName = ensure_numpy(m.columns)
#		m = wuml.ensure_numpy(m)
#
#		try: m = np.round(m, round_value)
#		except: pass
#		if cName is not None: m = np.vstack((cName,m))
#		m = str(m)
#	
#		if verticalize:
#			if len(m.shape) == 1:
#				m = np.atleast_2d(m).T
#
#		out_str = front_tab + str(m).replace('\n ','\n' + front_tab).replace('[[','[').replace(']]',']') + end_space + '\n'
#		out_str = str(out_str).replace('.]',']')
#
#		if wtype(title) == 'str':
#			L1 = out_str.split('\n')
#			L1_max_width = len(max(L1, key=len))
#			t1 = str.center(title, L1_max_width)
#			out_str = t1 + '\n' + out_str	
#	if auto_print: wuml.jupyter_print(out_str)
#	else: return out_str



def append_same_string_infront_of_block_of_string(front_string, back_block_string):

	new_str = front_string.join(back_block_string.splitlines(True))
	new_str = front_string + new_str
	return new_str


def block_two_string_concatenate(str1, str2, spacing='\t', add_titles=[], auto_print=False):
	str1 = str(str1)
	str2 = str(str2)

	L1 = str1.split('\n')
	#L2 = str2.strip().split('\n')
	L2 = str2.split('\n')

	if len(L1) > len(L2):
		Δ = len(L1) - len(L2)
		for ι in range(Δ):
			L2.append('\n')

	if len(add_titles) == 2:
		L1_max_width = len(max(L1, key=len))
		L2_max_width = len(max(L2, key=len))
		t1 = str.center(add_titles[0], L1_max_width)
		t2 = str.center(add_titles[1], L2_max_width)
		L1.insert(0,t1)
		L2.insert(0,t2)

	max_width = len(max(L1, key=len))
	outS = ''
	for l1, l2 in zip(L1,L2):
		outS += ('%-' + str(max_width) + 's' + spacing + l2 + '\n') % l1

	if auto_print: wuml.jupyter_print(outS, display_all_rows=True, display_all_columns=True)
	else: return outS


def block_matrix_concatenate(matrix_list, spacing='\t\t'):

	strList = []
	for m in matrix_list:
		#strList.append(pretty_np_array(m).strip())
		strList.append(pretty_np_array(m))

	return block_string_concatenate(strList, spacing=spacing)

def block_string_concatenate(strList, spacing='\t'):
	all_concat = ''
	for single_block in strList:

		if all_concat == '':
			all_concat = single_block
		else:
			all_concat = block_two_string_concatenate(all_concat, single_block, spacing=spacing)

	return all_concat

if __name__ == "__main__":
	str1 = 'Hello world\nA interesting string'
	str2 = 'Second Block\nto the right\nTooLong'
	outS = block_two_string_concatenate(str1, str2)
	print(outS)

	str1 = 'Hello world\nA interesting string'
	str2 = 'Second Block, to the right, TooLong'
	str3 = 'Third Block, \n make it shorter'
	all_3 = [str1, str2, str3]
	outS = block_string_concatenate(all_3)
	print(outS)





#!/usr/bin/env python

import itertools
from itertools import permutations
 

def zip_with_joint_permutation(*arg):
	return list(itertools.product(*arg))





if __name__ == "__main__":
	a = [1,2]
	b = ['a','b']
	e = ['s','t']

	O = zip_with_joint_permutation(a,b,e)
	for i, j, k in O: print(i,j, k) 

	import pdb; pdb.set_trace()



import numpy as np
import pandas as pd

import os
import sys
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')
import wuml

from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.naive_bayes import GaussianNB
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.inspection import permutation_importance
from sklearn.linear_model import LogisticRegression
import matplotlib.pyplot as plt
from scipy import stats


class classification:
	'''
	Automatically run classification on data

	data: can be any data type
	classifier='GP', 'SVM', 'RandomForest', 'KNN', 'NeuralNet', 'LDA', 'NaiveBayes', 'IKDR', 'LogisticRegression'
	split_train_test: automatically splits the data, default is true
	q: for IKDR, dimension to reduce to
	'''

	def __init__(self, data, y=None, test_data=None, testY=None, y_column_name=None, split_train_test=True, 	
				classifier='GP', kernel='rbf', 	# kernels = 'rbf', 'linear'
				reduce_dimension_first_method=None, # 'LDA'
				networkStructure=[(100,'relu'),(100,'relu'),(1,'none')], max_epoch=500, learning_rate=0.001, q=2,
				accuracy_rounding=3, regularization_weight=1):
		NP = wuml.ensure_numpy
		S = np.squeeze

		self.X_train = None
		self.data = data
		X = NP(data)
		if y is not None:
			y = S(NP(y))
		elif y_column_name is not None:
			y = data[y_column_name].values
		elif type(data).__name__ == 'wData':
			y = data.Y
		else: raise ValueError('Undefined label Y')

		if test_data is not None:
			if wuml.wtype(data) == 'wData': self.X_train = data.X
			elif wuml.wtype(data) == 'ndarray': self.X_train = data

			if y is None: self.y_train = data.Y
			else: self.y_train = y

			if wuml.wtype(test_data) == 'wData': self.X_test = test_data.X
			elif wuml.wtype(test_data) == 'ndarray': self.X_test = test_data
			
			if testY is None: self.y_test = test_data.Y
			else: self.y_test = testY
			split_train_test = True
		else:
			if split_train_test:
				self.X_train, self.X_test, self.y_train, self.y_test = wuml.split_training_test(data, label=y, xdata_type="%.4f", ydata_type="%.4f")
			else:
				self.X_train = NP(X)
				self.y_train = NP(y)


		#Check for dimension reduction, example LDA+SVM
		cfname = classifier.split('+')
		self.original_classifier_name = classifier
		if len(cfname) == 2:
			reduce_dimension_first_method = cfname[0]
			classifier = cfname[1]

		# Reduce the dimension of data first if LDA is required
		if reduce_dimension_first_method == 'LDA':
			self.dim_reduct = LinearDiscriminantAnalysis()
			self.dim_reduct.fit(NP(self.X_train), S(NP(self.y_train)))
			self.X_train = NP(self.X_train).dot(self.dim_reduct.coef_.T)
			if split_train_test: self.X_test = NP(self.X_test).dot(self.dim_reduct.coef_.T)


		if classifier == 'GP':
			if kernel == 'rbf': kernel = 1.0 * RBF(1.0) 
			model = GaussianProcessClassifier(kernel=kernel, random_state=0)
		elif classifier == 'SVM':
			model = make_pipeline(StandardScaler(), SVC(gamma='auto', kernel=kernel, C=regularization_weight))
		elif classifier == 'RandomForest':
			model = RandomForestClassifier(max_depth=3, random_state=0)
		elif classifier == 'KNN':
			model = KNeighborsClassifier(n_neighbors=4)
		elif classifier == 'NeuralNet':
			model = MLPClassifier(random_state=1, max_iter=max_epoch)
		elif classifier == 'LDA':
			model = LinearDiscriminantAnalysis()
		elif classifier == 'NaiveBayes':
			model = GaussianNB()
		elif classifier == 'GradientBoosting':
			model = GradientBoostingClassifier(n_estimators=100, learning_rate=1.0, max_depth=1, random_state=0)
		elif classifier == 'LogisticRegression':
			model = LogisticRegression(random_state=0)
		elif classifier == 'IKDR':
			model = wuml.IKDR(self.X_train, q=q, y=self.y_train)
		else: raise ValueError('Unrecognized Classifier')

		NPR = np.round
		
		model.fit(NP(self.X_train), S(NP(self.y_train)))
		self.ŷ_train = model.predict(NP(self.X_train))
		self.Train_acc = NPR(wuml.accuracy(S(NP(self.y_train)), self.ŷ_train), accuracy_rounding)

		if split_train_test:
			self.ŷ_test = model.predict(NP(self.X_test))
			self.Test_acc = NPR(wuml.accuracy(S(NP(self.y_test)), self.ŷ_test), accuracy_rounding)

		#import pdb; pdb.set_trace()
		self.split_train_test = split_train_test
		self.model = model
		self.classifier = classifier
		self.kernel = kernel
		self.reduce_dimension_first_method = reduce_dimension_first_method
		#self.results = self.result_summary(print_out=False)

	def project_data_onto_linear_weights(self, X=None):
		# if X is none, it will use the original data
		if wuml.wtype(self.model) != 'LinearDiscriminantAnalysis':
			print('Warning: the function project_data_onto_linear_weights only works for LDA')
			return X

		if X is None:
			nX = wuml.ensure_numpy(self.data)
			pX = nX.dot(self.model.coef_.T)

		return pX
			
	def output_sorted_feature_importance_table(self, column_names=None, show_top_few=5): 	# feature importance computed via permutation_importance
		NP = wuml.ensure_numpy

		all_classifiers =['GP', 'SVM', 'RandomForest', 'KNN', 'NeuralNet', 'LDA', 'NaiveBayes', 'IKDR', 'LogisticRegression']

		if column_names is None:
			Cnames = wuml.ensure_wData(self.data).columns
		else:
			Cnames = wuml.ensure_list(column_names)

		importance_GP = permutation_importance(self.model, NP(self.X_train), self.y_train, scoring='accuracy')
		importance = importance_GP.importances_mean

		if self.classifier in all_classifiers:

			coefs = pd.DataFrame( importance, columns=['Coefficients'], index=Cnames)
			sorted_coefs = coefs.sort_values(by='Coefficients', ascending=False)
			first_few = sorted_coefs.head(show_top_few)
			wuml.jupyter_print(first_few)

	def plot_feature_importance(self, title, Column_names, title_fontsize=12, axis_fontsize=9, xticker_rotate=0, ticker_fontsize=9,
								yticker_rotate=0, ytick_locations=None, ytick_labels=None): # feature importance computed via permutation_importance

		if wuml.get_commandLine_input()[1] == 'disabled': return
		NP = wuml.ensure_numpy
		all_classifiers =['GP', 'SVM', 'RandomForest', 'KNN', 'NeuralNet', 'LDA', 'NaiveBayes', 'IKDR']
		Cnames = wuml.ensure_list(Column_names)

		importance_GP = permutation_importance(self.model, NP(self.X_train), self.y_train, scoring='accuracy')
		importance = importance_GP.importances_mean

		if self.classifier in all_classifiers:

			coefs = pd.DataFrame( importance, columns=['Coefficients'], index=Cnames)
				
			#coefs.plot(kind='barh', figsize=(9, 7))
			coefs.plot(kind='barh')
			plt.title(title, fontsize=title_fontsize)
			plt.axvline(x=0, color='.5')
			plt.subplots_adjust(left=.3)
			plt.tight_layout()
			plt.ylabel('Features', fontsize=axis_fontsize)
			plt.xlabel('Features Influence on Results', fontsize=axis_fontsize)
			plt.yticks(fontsize=ticker_fontsize, rotation=yticker_rotate, ticks=ytick_locations, labels=ytick_labels )

			plt.show()

	def result_summary(self, print_out=True):
		if self.split_train_test:
			column_names = ['classifier', 'Train', 'Test']
			data = np.array([[self.original_classifier_name, self.Train_acc , self.Test_acc]])
		else:
			column_names = ['classifier', 'Train']
			data = np.array([[self.original_classifier_name, self.Train_acc]])

		df = pd.DataFrame(data, columns=column_names,index=[''])
		if print_out: wuml.jupyter_print('\n', df)
		self.results = df
		return df


	def save_classifier_to_pickle_file(self, path):
		wuml.pickle_dump(self, path)

	def __call__(self, data):

		X = wuml.ensure_numpy(data)	
		if self.reduce_dimension_first_method == 'LDA':
			X = X.dot(self.dim_reduct.coef_.T)

		try: [self.ŷ, self.σ] = self.model.predict(X, return_std=True, return_cov=False)
		except: self.ŷ = self.model.predict(X)

		self.ŷ = wuml.ensure_data_type(self.ŷ, type_name=type(data).__name__)
		try: return [self.ŷ, self.σ]
		except: return self.ŷ

		

	def __str__(self):
		#return self.results
		return str(self.result_summary(print_out=False))


def run_every_classifier(data, y=None, y_column_name=None, order_by='Test', q=None, kernel='rbf',
						classifiers=['GP', 'SVM', 'RandomForest', 'KNN', 'NeuralNet', 'LDA', 'NaiveBayes', 'IKDR','LogisticRegression']	):
	'''
	data is type wData
	order_by: 'Test', 'Train'
	q: for ikdr
	'''
	
	results = {}
	
	if q is None: q = int(data.shape[1]/2)
	df = pd.DataFrame()
	for reg in classifiers:
		wuml.write_to_current_line('Running %s'%reg)
		results[reg] = classification(data, y=y, classifier=reg, split_train_test=True, q=q, kernel='rbf')
		df = df.append(results[reg].result_summary(print_out=False))

	if order_by == 'Test': results['Train/Test Summary'] = df.sort_values(['Test','Train'], ascending=False)
	else: results['Train/Test Summary'] = df.sort_values(['Train','Test'], ascending=False)

	return results


#	Generate 10 classifiers and use bagging concensus to reach a label
class ten_folder_classifier:
	def __init__(self, data, y=None, y_column_name=None, q=2, kernel='linear', classifier='SVM'):

		tenFoldData = wuml.gen_10_fold_data(data=data)
		self.tb = tb = wuml.result_table(column_names=['Fold', 'Train Acc', 'Test Acc'])
	
		self.classifier_list = []
		for i, fold in enumerate(tenFoldData):
			wuml.write_to_current_line('Running fold :%d'%(i+1))
			[X_train, Y_train, X_test, Y_test] = fold
			cf = wuml.classification(X_train, y=Y_train, test_data=X_test, testY=Y_test, classifier=classifier, kernel=kernel, q=q)
			self.classifier_list.append(cf)
			tb.add_row([i+1, cf.Train_acc , cf.Test_acc])
	
		TrainAcc = np.mean(tb.get_column('Train Acc').X)
		TestAcc = np.mean(tb.get_column('Test Acc').X)
		tb.add_row(['Avg Acc', TrainAcc , TestAcc])

	def show_results(self):
		wuml.jupyter_print(self.tb)

	def predict(self, data):
		return self.__call__(data)

	def __call__(self, data):
		all_labels = None
		for model in self.classifier_list:
			if all_labels is None:
				all_labels = wuml.ensure_wData(model(data))
			else:
				label = wuml.ensure_wData(model(data))
				all_labels.append_columns(model(data))

		m = stats.mode(all_labels.X, axis=1)
		return m[0]

	def fit(self, data, y=None):
		tenFoldData = wuml.gen_10_fold_data(data=data)
	
		self.classifier_list = []
		for i, fold in enumerate(tenFoldData):
			wuml.write_to_current_line('Running fold :%d'%(i+1))
			[X_train, Y_train, X_test, Y_test] = fold
			cf = wuml.classification(X_train, y=Y_train, test_data=X_test, testY=Y_test, classifier=classifier, kernel=kernel, q=q)
			self.classifier_list.append(cf)

	def save_classifier_to_pickle_file(self, path):
		wuml.pickle_dump(self, path)





import sys
from wuml.basicNetwork import *
from wuml.type_check import *
import torch
import numpy as np


class combinedNetwork():
	def __init__(self, data, netStructureList, netInputDimList, costFunction, network_behavior_on_call, optimizer_steps_order=None, 
						early_exit_loss_threshold=0.0000001, 
						on_new_epoch_call_back = None, pickled_network_info=None, explainer=None,
						max_epoch=1000, data_mean=None, data_std=None,	force_network_to_use_CPU=False,
						X_dataType=torch.FloatTensor, Y_dataType=torch.FloatTensor, extra_dataType=None,
						network_usage_output_type='Tensor', learning_rate=0.001, lr_decay_rate=0.5, lr_decay_patience=50):
		'''
			data: must be a wData
			netStructureList: must be a list, can have multiple networks all working together
			on_new_epoch_call_back: end of each epoch calls this function
			lr_decay_rate: the current lr multiply by this value to decay, 1 would be no decay
			lr_decay_patience: the number of epochs which sees no improvement then trigger decay
			network_behavior_on_call:  what happens if you call  net(data) after instantiation
		'''
		if get_commandLine_input()[1] == 'disabled': max_epoch = 10
		if force_network_to_use_CPU: wuml.pytorch_device = 'cpu'
		self.explainer = explainer			# If this model's features were explained, we save the explainer here

		# the network is awared if the data was centered
		if wtype(data) is 'wData':
			self.μ = data.μ
			self.σ = data.σ
		else:
			self.μ = data_mean
			self.σ = data_std


		self.network_behavior_on_call = network_behavior_on_call
		self.on_new_epoch_call_back = on_new_epoch_call_back #set this as a callback at each function
		self.costFunction = costFunction
		self.device = wuml.get_current_device()

		if pickled_network_info is None:
			dat = wuml.ensure_wData(data)
	
			self.batch_size = data.batch_size
			self.netStructureList = netStructureList
			self.netInputDimList = netInputDimList
			self.optimizer_steps_order = optimizer_steps_order
			self.networkList = []
	
			self.X_dataType = X_dataType
			self.Y_dataType = Y_dataType
			self.extra_dataType=extra_dataType
			if wtype(extra_dataType) != 'list' and extra_dataType is not None:
				self.extra_dataType = [extra_dataType]
	
			# double check that extra data is properly defined
			if dat.extra_data_dictionary['extra_data'] is not None:
				if extra_dataType is None:
					ValueError('Error: If you have extra data within input, define the extra_dataType parameter as a list of data types.')
				if len(dat.extra_data_dictionary['extra_data']) != len(extra_dataType):
					raise ValueError('Error: If you have extra data within input, you must have a list of the same length for extra_dataType, e.g., extra_dataType=[torch.FloatTensor, torch.LongTensor]')
	
	
			self.trainLoader = dat.get_data_as('DataLoader')
			self.lr = learning_rate
			self.lr_decay_patience = lr_decay_patience
			self.lr_decay_rate = lr_decay_rate
			self.max_epoch = max_epoch
			self.early_exit_loss_threshold = early_exit_loss_threshold
	

			for NetStruct, dim in zip(netStructureList, netInputDimList):
				newNet = flexable_Model(dim, NetStruct)
				newNet.to(self.device)		# store the network weights in gpu or cpu device
				#newNet = basicNetwork(costFunction, dat, networkStructure=NetStruct, network_info_print=False, override_network_input_width_as=dim, max_epoch=self.max_epoch)
				self.networkList.append(newNet)

		else:
			self.lr = pickled_network_info['lr']
			self.max_epoch = pickled_network_info['max_epoch']
			self.netStructureList = pickled_network_info['netStructureList'] 
			self.networkList = pickled_network_info['networkList']
			self.extra_dataType = pickled_network_info['extra_dataType']
			self.batch_size = 32

			for model in self.networkList:
				model.to(self.device)
			

		self.info()

	def center_and_scaled(self, data):
		# the network is awared if the data was centered
		if self.μ is not None and self.σ is not None:
			X = ensure_numpy(data)
			X = (X - self.μ)/self.σ
			try:
				return ensure_data_type(X, type_name=wtype(data), ensure_column_format=True, column_names=data.columns)
			except:
				return ensure_data_type(X, type_name=wtype(data), ensure_column_format=True)
		else:
			return data

	def output_network_data_for_storage(self):
		net = {}
		net['name'] = self.__class__.__name__
		net['network_behavior_on_call'] = marshal.dumps(self.network_behavior_on_call.__code__)
		net['costFunction'] = marshal.dumps(self.costFunction.__code__)

		#if self.on_new_epoch_call_back is None: net['on_new_epoch_call_back'] = None
		#else: net['on_new_epoch_call_back'] = marshal.dumps(self.on_new_epoch_call_back.__code__)

		net['lr'] = self.lr
		net['max_epoch'] = self.max_epoch
		net['netStructureList'] = self.netStructureList
		net['networkList'] = self.networkList
		net['extra_dataType'] = self.extra_dataType
		net['μ'] = self.μ
		net['σ'] = self.σ
		if wtype(self.explainer) == 'explainer':
			net = self.explainer.output_network_data_for_storage(net)

		return net


	def info(self, printOut=True):
		info_str = 'All Networks\n' 
		info_str += '\tBatch size: %.d: \n'%self.batch_size
		info_str += '\tLearning rate: %.3f\n'%self.lr
		info_str += '\tMax number of epochs: %d\n'%self.max_epoch
		info_str += '\tCost Function: %s\n'%wuml.get_function_name(self.costFunction)
		if self.on_new_epoch_call_back is not None: info_str += '\tTrain Loop Callback: %s\n'%wuml.get_function_name(self.on_new_epoch_call_back)
		info_str += '\tDevice type: %r\n'%self.device

		for j, net in enumerate(self.networkList):
			info_str += '\tNetworks %d Structure\n'%(j) 

			for i in net.children():
				try: info_str += ('\t\t%s , %s\n'%(i,i.activation))
				except: info_str += ('\t\t%s \n'%(i))

		if printOut: wuml.jupyter_print(info_str)
		return info_str


	def fit(self, print_status=True):
		netParams = []
		netSchedulers = []
		netOptimizers = []

		for net in self.networkList:
			params = net.parameters()
			optimizer = torch.optim.Adam(params, lr=self.lr)
			scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau( optimizer, factor=self.lr_decay_rate, min_lr=1e-10, patience=50, verbose=False)	
					# factor is the decay rate
					# patience is how many epoch until trigger decay

			netParams.append(params)
			netSchedulers.append(scheduler)
			netOptimizers.append(optimizer)

		for epoch in range(self.max_epoch):
	
			loss_history = []	
			for (i, data) in enumerate(self.trainLoader):
				x = Variable(data[0].type(self.X_dataType), requires_grad=False)
				y = Variable(data[1].type(self.Y_dataType), requires_grad=False)
				x= x.to(self.device, non_blocking=True )
				y= y.to(self.device, non_blocking=True )
				formatted_data = [x, y, data[2]]
			
				for j, dtype in enumerate(self.extra_dataType): 
					extraDat = Variable(data[j+3].type(dtype), requires_grad=False)
					extraDat= extraDat.to(self.device, non_blocking=True )
					formatted_data.append(extraDat)

				for opt in netOptimizers: opt.zero_grad()
				all_losses = self.costFunction(formatted_data, self.networkList)	

				if wtype(all_losses) == 'list': main_loss = all_losses[0]		# all_losses is a list of losses, but the 1st one is always the main loss
				else: main_loss = all_losses

				main_loss.backward()
				if self.optimizer_steps_order is None:
					for opt in netOptimizers: opt.step()
				else:
					self.optimizer_steps_order(netOptimizers)

				loss_history.append(main_loss.item())
	
			loss_avg = np.array(loss_history).mean()
			for sched in netSchedulers: sched.step(loss_avg)
			if loss_avg < self.early_exit_loss_threshold: break;


			#	The default status printing is just the main loss, but if you return a list of loss, 
			#	you can print your own style by adding the function on_new_epoch_call_back()
			if print_status:
				if self.on_new_epoch_call_back is not None:
					self.on_new_epoch_call_back(all_losses, (epoch+1), netSchedulers[0]._last_lr[0])
				else:
					if get_commandLine_input()[1] != 'disabled': 
						txt = '\tepoch: %d, Avg Loss/dimension: %.4f, Learning Rate: %.8f'%((epoch+1), loss_avg, enc_scheduler._last_lr[0])
						write_to_current_line(txt)
		
		for net in self.networkList: net.eval()

	def __call__(self, data, output_type='Tensor', out_structural=None):
		'''
			out_structural (mostly for classification purpose): None, '1d_labels', 'one_hot'
		'''

		data = ensure_wData(data)
		X = ensure_proper_model_input_format(data)
		X = ensure_tensor(X, dataType=torch.FloatTensor)
	
		if data.Y is None: formatted_data = [X,None]
		else:
			y = ensure_tensor(data.Y, dataType=torch.FloatTensor)
			formatted_data = [X,y]


		if wtype(data) == 'wData':
			extra_data = data.extra_data_dictionary['extra_data']
			if extra_data is not None:
				for j, xData in enumerate(extra_data): 
					newD = ensure_tensor(xData, dataType=self.extra_dataType[j])
					formatted_data.append(newD)

		all_net_output = self.network_behavior_on_call(formatted_data, self.networkList)	

		if wtype(all_net_output) == 'list':
			return wuml.cast_each_item_in_list_as(all_net_output, output_type)
		elif wtype(all_net_output) == 'Tensor':
			return ensure_data_type(all_net_output, type_name=output_type)
		else:
			return all_net_output
			








import os
import sys
if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')
if os.path.exists('/home/chieh/code/interpret'):
	sys.path.insert(0,'/home/chieh/code/interpret')

import wuml
from wuml.type_check import *
from wuml.IO import *
import numpy as np
import pandas as pd
import marshal, types

from interpret import set_visualize_provider
from interpret.provider import InlineProvider
set_visualize_provider(InlineProvider())

from interpret import show
from interpret.blackbox import LimeTabular
from interpret.blackbox import ShapKernel

class explainer():
	def __init__(self, data, model, explainer_algorithm='shap', which_model_output_to_use=None):
		'''
			data : need to be in wData
			which_model_output_to_use: sometimes the model output multiple outputs in a list instead of just the label, 
					this integer tells us which item within the last should be used
		'''
		self.which_model_output_to_use = which_model_output_to_use
		self.data = wuml.ensure_wData(data)
		self.explainer_algorithm = explainer_algorithm
		self.column_names = self.data.column_names

		reduce_down_to = 70
		if self.data.shape[0] > reduce_down_to:
			km = wuml.clustering(self.data, n_clusters=reduce_down_to, method='KMeans')
			Cs = km.model.cluster_centers_
			self.data = wuml.ensure_wData(Cs, column_names=data.columns)
			jupyter_print('\nNote: Since there are too many input samples, kmeans was employed to reduce the sample down to %d\n'%reduce_down_to)

		self.joint_table = None		# initialize a table of all feature importances
		if wtype(model) == 'autoencoder': self.model_prediction = model.objective_network
		else: self.model_prediction = model
		

		explain_types = {'lime':LimeTabular, 'shap':ShapKernel}
		self.explainer = explain_types[explainer_algorithm](predict_fn=self.model_predict_wrapper, data=self.data.df)
		model.explainer = self


	def output_network_data_for_storage(self, existing_db=None):
		if existing_db is None: existing_db = {}

		existing_db['explainer'] = 'exists'
		existing_db['reference_data'] = self.data.X
		existing_db['explainer_algorithm'] = self.explainer_algorithm
		existing_db['which_model_output_to_use'] = self.which_model_output_to_use
		existing_db['column_names'] = self.column_names

		return existing_db

	def model_predict_wrapper(self, X_input):
		y = self.model_prediction(X_input)
		if self.which_model_output_to_use is not None:
			y = y[self.which_model_output_to_use]

		y = ensure_numpy(y, ensure_column_format=False)
		y = np.squeeze(y)

		if len(y.shape) == 0: y = np.array([y])
		return y



	def __call__(self, data, y=None, display=True, outpath=None, figsize=None):
		data = wuml.ensure_wData(data)
		if y is not None:
			local_explain = self.explainer.explain_local(data.df, y)
		else:
			local_explain = self.explainer.explain_local(data.df, data.Y)

		if display: show(local_explain)

		self.joint_table = self.feature_importance_table(local_explain, print_out=display)
		self.global_summary = self.most_important_features(local_explain, print_out=display)

		#if figsize is not None or outpath is not None:
		#	# plot out and save importance
		#	[y, ŷ, Δy] = ensure_list(np.round(self.joint_table.X, 4)[0])[-3:]
		#	pred_str = 'y = %.3f\nŷ = %.3f'%(y, ŷ)
		#	for i in range(self.joint_table.shape[0]): 
		#		feature_importance = self.joint_table.X[i]
		#		feature_importance = feature_importance[0:len(feature_importance)-3]
		#		B = wplotlib.bar(self.column_names, feature_importance, 'Sample %s: '%i, 'Features', 'Importance', horizontal=True, imgText=pred_str, xTextShift=1.05, yTextShift=0, figsize=figsize, outpath=output)

		return [self.joint_table, self.global_summary]

	def feature_importance_table(self, local_explain, print_out=True):
		E = local_explain._internal_obj['specific']
		explain_list = []
		for local_explain in E:
			α = local_explain['perf']['actual_score']
			β = local_explain['perf']['predicted_score']
			γ = np.absolute(α - β)
			σ = np.append(local_explain['scores'], [α, β, γ])
			n = np.append(local_explain['names'], ['y', 'ŷ', 'Δy'])

			item = pd.DataFrame(np.array([σ]), columns=n)
			explain_list.append(item)

		cnames = np.append(self.column_names, ['y', 'ŷ', 'Δy'])
		result = pd.concat(explain_list)
		r2 = result.reindex(columns=cnames)
		self.joint_table = ensure_wData(r2.reset_index(drop=True))

		if print_out: jupyter_print(self.joint_table, endString='\n', display_all_rows=True, display_all_columns=True)
		return self.joint_table

	def most_important_features(self, local_explain, print_out=True):
		E = local_explain._internal_obj['specific']
		feature_names = E[0]['names']
		D = {}
		Dm = {}
		for j in feature_names: 
			D[j] = 0
			Dm[j] = 0

		for i in E:
			most_important_feature = i['names'][np.argmax(np.absolute(i['scores']))]
			D[most_important_feature] += 1		#increase count

			for j, k in enumerate(i['names']): 
				Dm[k] += np.absolute(i['scores'][j])

		D = wuml.ensure_wData(pd.DataFrame(D.values(),index=D.keys(), columns=['Most chosen']))
		Dm = wuml.ensure_wData(pd.DataFrame(Dm.values(),index=Dm.keys(), columns=['Most weighted']))
		D.sort_by('Most chosen', ascending=False)
		Dm.sort_by('Most weighted', ascending=False)
		if print_out: wuml.print_two_matrices_side_by_side(D.df, Dm.df, title1='', title2='')
	
		return [D, Dm]
		

	def plot_individual_sample_importance(self, one_sample, y=None, sample_id='', figsize=None):
		X = ensure_proper_model_input_format(one_sample)
		[self.joint_table, self.global_summary] = self.__call__(X, y=y, display=False)

		cnames = ensure_list(self.column_names)
		[y, ŷ, Δy] = ensure_list(np.round(self.joint_table.X, 4)[0])[-3:]
		pred_str = 'y = %.3f\nŷ = %.3f'%(y, ŷ)
		feature_importance = ensure_list(ensure_numpy(self.joint_table).round(4))[0]
		feature_importance = feature_importance[0:len(feature_importance)-3]		# remove the last 3 elements cus they are ['y', 'ŷ', 'Δy']
		B = wplotlib.bar(cnames, feature_importance, 'Sample %s: '%sample_id, 'Features', 'Importance', horizontal=True, imgText=pred_str, xTextShift=1.05, yTextShift=0, figsize=figsize)	



#!/usr/bin/env python
# -*- coding: utf-8 -*-


#	Written by Chieh Wu, 9/13/2022
#	This function calculates the Gaussian Kernel by approximate it through Random fourier Feature and Orthogonal Random Feature technique.

#	**Random Fourier Feature**
#	Given data $X \in \mathbb{R}^{n \times d}$, RFF says that a kernel function is 
#	$$k(x,y) = E_{p(\omega)}[ cos(w^T x + \theta) cos(w^T y + \theta)] $$
#	and therefore can be approximated with
#	$$k(x,y) = \frac{2}{m} \sum_i^m cos(w_i^T x + \theta) cos(w_i^T y + \theta) $$
#	This implies that the feature map is 
#	$$ \phi(x) \approx \sqrt{\frac{2}{m}} [cos(w_1^T x + \theta), cos(w_2^T x + \theta), ... , cos(w_m^T x + \theta)]$$
#	Given $W \in \mathbb{R}^{d \times m}$ the entire dataset $X$, then
#	$$ \phi(X) \approx \sqrt{\frac{2}{m}} cos[XW + \theta]$$
#	The the element of the W matrix is a normal gaussian distribution divided by σ of the gaussian distribution
#	θ is randomly generated using uniform distribution between 0 to 2π


#	**Structured Orthogonal Random Feature SORF**
#	The SORF replaces the $W$ matrix with
#	$$W = \frac{\sqrt{d}}{\sigma} HD_1HD_2HD_3$$
#	$H$ is a normalized Hadamard matrix, the normalization is such that $I = H^TH$
#	Therefore $H$ is the Hadamard matrix divided by $\frac{1}{\sqrt{d}}$
#	and D is a diagonal matrix with the diagonal elements sampled from Rademacher Distribution

#	Note that since Hadamard matrix size is alway in the power of 2 $2^n$, the dimension of the data is probably not the same.
#	In these cases, pad the data with 0s as the extra dimensions. Also, the width of the SORF is always set to the dimension
#	of the data after padding. Therefore, to make the width even longer, we must generate multiple size d matrices and 
#	concatenate them together. 

import os 
import sys

if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')

import numpy as np
import sklearn.metrics
import torch
import numpy.matlib
from torch.autograd import Variable
from numpy.random import rand
import torch.nn.functional as F
from scipy.linalg import hadamard
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
import wuml
from wuml import jupyter_print

class random_feature:
	# sample_num, the larger the better approximation
	# random_feature_method can be 'orthogonal' (default) or 'rff'
	def __init__(self, kernel='rbk', sigma=1, random_feature_method='orthogonal', sample_num=500):
		self.method = random_feature_method
		self.kernel = kernel
		self.m = sample_num		# the number of samples used to approximate k(x,y)
		self.θ = None			# phase shift
		self.σ = sigma
#
	def sample_rademacher_distribution(self, num_of_samples):
		r = np.random.binomial(1, 0.5, int(num_of_samples))
		r = 2*(r - 0.5)
		D = np.diag(r)
		return D
#
	def initialize_random_features(self, X):
		if self.θ == None:
			X0 = X
			N = X.shape[0]
			d = X.shape[1]
			σˉᑊ = 1/self.σ
			ƻπ = 2*np.pi
			m = self.m
			dᑊᐟᒾ = np.sqrt(d)
	
			if self.method == 'orthogonal':	# perform SORF
				#	Here we will pad the data to ensure data dimension is in power of 2 for Hadamard matrix
				ḿ = np.power(2, np.ceil(np.log2(d))) 	# dimension in power of 2 just greater than d
				Δ = int(ḿ - d)
				if Δ != 0:
					pad0 = np.zeros((N, Δ))
					X0 = np.concatenate((X, pad0), axis=1)
				#	each loop creates ḿ out of m, we need to Figure out how many times we need to regenerate 
				#	Note the sample_num may not be equal to the actual random feature width M
				#		m = the number of random features width suggested by the user
				#		ḿ = the number of random features generated by each hadamard matrix 
				#		M = the acutal number of random features we actually use since power of 2 is required by hadamard
				repeat = int(np.ceil(m/ḿ))
				W = np.empty((int(ḿ),0))
				M = int(repeat*ḿ)
				for ɷ in range(repeat):
					#	Generate H and diagonal matrix D sampled from rademacher distribution
					H = 1/np.sqrt(ḿ)*hadamard(ḿ)		# normalized H to be orthonormal
					Dᑊ = self.sample_rademacher_distribution(ḿ)
					Dᒾ = self.sample_rademacher_distribution(ḿ)
					Dᶾ = self.sample_rademacher_distribution(ḿ)
					#
					HDᑊHDᒾHDᶾ = H.dot(Dᑊ).dot(H).dot(Dᒾ).dot(H).dot(Dᶾ)
					Ŵ = (dᑊᐟᒾ*σˉᑊ*HDᑊHDᒾHDᶾ).T			# we did follow the paper of Wx and use XW format
					W = np.concatenate((W, Ŵ), axis=1)
					#
				θ = ƻπ*rand(1, M)
				return [W, X0, M, θ]
				#
			elif self.method == 'rff':
				θ = ƻπ*rand(1, m)
				W = (σˉᑊ)*np.random.randn(d, m)	# random projection matrix W
				return [W, X, m, θ]
#
	def get_feature_map(self, X):
		#	input data conversion
		input_data_type = wuml.wtype(X)
		X = wuml.ensure_data_type(X)

		[W,Ẋ, m, θ] = self.initialize_random_features(X)
		#
		ẊW = Ẋ.dot(W)
		c = np.sqrt(2.0/m)	
		Φ = c*np.cos(ẊW + θ)	

		#	output data conversion
		Φ = wuml.ensure_data_type(Φ, type_name=input_data_type)

		return Φ
#
	def get_kernel(self, X):
		#	input data conversion
		input_data_type = wuml.wtype(X)
		X = wuml.ensure_data_type(X)

		Φ = self.get_feature_map(X)
		K = Φ.dot(Φ.T)
		K = np.maximum(0,K)	# run a relu on the kernel so no negative values
		if self.kernel == 'rbk': K = np.minimum(1,K) # make sure the kernel values doesn't go beyond 1 for gaussian

		#	output data conversion
		K = wuml.ensure_data_type(K, type_name=input_data_type)
		return K

if __name__ == "__main__":
	np.set_printoptions(precision=3)
	np.set_printoptions(linewidth=300)
	np.set_printoptions(suppress=True)
#
	#	On lower dimension data, it seems that RFF is good enough
	X = np.random.randn(7,5)
	σ = np.median(sklearn.metrics.pairwise.pairwise_distances(X))
	γ = 1.0/(2*σ*σ)
#
	K = sklearn.metrics.pairwise.rbf_kernel(X, gamma=γ)
#
	sorf = random_feature(sigma=σ, random_feature_method='orthogonal')
	Kₒ = sorf.get_kernel(X)			# kernel matrix from orthogonal
#
	rff = random_feature(sigma=σ, random_feature_method='rff')
	Kᵣ = rff.get_kernel(X)			# kernel matrix from orthogonal
#
#
	wuml.print_two_matrices_side_by_side(K, Kₒ, title1='Real', title2='Approx by SORF', auto_print=True)
	wuml.print_two_matrices_side_by_side(K, Kᵣ, title1='Real', title2='Approx by RFF', auto_print=True)
#
	jupyter_print('Notice that RFFperforms better on lower dimension datasets')
	ε = mean_absolute_error(K, Kₒ)
	jupyter_print('Mean absolute Error with SORF %.3f'%ε)
	ε = mean_absolute_error(K, Kᵣ)
	jupyter_print('Mean absolute Error with RFF %.3f\n'%ε)
#
	ε = mean_squared_error(K, Kₒ)
	jupyter_print('Mean Squared Error with SORF %.3f'%ε)
	ε = mean_squared_error(K, Kᵣ)
	jupyter_print('Mean Squared Error with SORF %.3f'%ε)
#
#
	#	SORF starts working better when the dimension gets larger, here we use 16 dimension
	X = wuml.csv_load('../dataset/letters.csv', shuffle_samples=True)
	X = X[0:600,:]
#
	σ = np.median(sklearn.metrics.pairwise.pairwise_distances(X))
	γ = 1.0/(2*σ*σ)
#
	K = sklearn.metrics.pairwise.rbf_kernel(X, gamma=γ)
#
	sorf = random_feature(sigma=σ, random_feature_method='orthogonal')
	Kₒ = sorf.get_kernel(X)			# kernel matrix from orthogonal
#
	rff = random_feature(sigma=σ, random_feature_method='rff')
	Kᵣ = rff.get_kernel(X)			# kernel matrix from orthogonal
#
	jupyter_print('Notice that SORF performs better on higher dimension datasets')
	ε = mean_absolute_error(K, Kₒ)
	jupyter_print('\tMean absolute Error with SORF %.3f'%ε)
	ε = mean_absolute_error(K, Kᵣ)
	jupyter_print('\tMean absolute Error with RFF %.3f\n'%ε)
#
	ε = mean_squared_error(K, Kₒ)
	jupyter_print('\tMean Squared Error with SORF %.3f'%ε)
	ε = mean_squared_error(K, Kᵣ)
	jupyter_print('\tMean Squared Error with SORF %.3f'%ε)





import numpy as np
import random
import string


def gen_squared_symmetric_matrix(length, distribution='Gaussian'):
	if distribution == 'Gaussian':
		X = np.random.randn(length, length)
	elif distribution == 'Uniform':
		X = np.random.rand(length, length)

	X = X.dot(X.T)
	return X

def gen_random_string(str_len=5, letters=string.ascii_letters):
	#letters = string.ascii_lowercase
	#letters = string.ascii_letters
	#letters = string.ascii_uppercase
	#letters = string.digits
	#letters = string.punctuation
	return ''.join(random.choice(letters) for i in range(str_len)) 

def gen_exponential(λ=1, size=1):
	# λ exp(-λx) , μ=1/λ
	return np.random.exponential(scale=1/λ, size=size)


#category_range	: if 5, it will select from 0,1,2,3,4
#n				: number of samples returned	
#probabilities 	: the probability of each instance
#example		: np.random.choice(5, 3, p=[0.1, 0, 0.3, 0.6, 0])
def gen_categorical(category_range, n, probabilities):
	return np.random.choice(category_range, n, p=probabilities)
	

def sampling_rows_from_matrix(sample_size, matrix):
	number_of_rows = matrix.shape[0]
	random_indices = np.random.choice(number_of_rows, size=sample_size, replace=False)
	random_rows = matrix[random_indices, :]
	return random_rows




#!/usr/bin/env python3

import torch
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader
from sklearn import preprocessing
#from src.tools.path_tools import *
#from distances import *
#from src.tools.terminal_print import *
import numpy as np

class DManager(Dataset):
	def __init__(self, X, Y=None, extra_data_dictionary=None): 
	# X: data, Y: label, extra_data_dictionary: dictionary of extra data for anything

		self.X = X
		self.N = self.X.shape[0]
		self.d = self.X.shape[1]
		self.X_Var = torch.tensor(self.X, requires_grad=False)

		self.Y = Y
		if Y is not None: self.Y_Var = torch.tensor(self.Y, requires_grad=False)
		self.extra_data_dictionary = extra_data_dictionary


	def __getitem__(self, index):

		if self.extra_data_dictionary is None:
			if self.Y is None: items = [self.X[index], 0, index]
			else: items = [self.X[index], self.Y[index], index]
		else:
			if self.Y is None: Yout = 0
			else: Yout = self.Y[index]
			items = [self.X[index], Yout, index]
			for dat in self.extra_data_dictionary['numpy']:
				items.append(dat[index])

		return items

	def __len__(self):
		try: return self.X.shape[0]
		except: return 0





import os
import sys
if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')


from sklearn import preprocessing
import wuml

from wuml.type_check import *
import pandas as pd
import torch
import numpy as np
from sklearn.preprocessing import LabelEncoder
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader


class wData:
	def __init__(self, xpath=None, ypath=None, column_names=None, path_prefix='',
					label_column_name=None, label_column_id=None,
					dataFrame=None, X_npArray=None, Y_npArray=None, 
					extra_data=None, extra_data_preprocessing=None,
					first_row_is_label=False, 
					row_id_with_feature_names=None, first_column_as_sample_index=False, 
					label_type=None, label2_type=None, #it should be either 'continuous' or 'discrete'
					encode_discrete_label_to_one_hot=False,
					xtorchDataType=torch.FloatTensor, ytorchDataType=torch.FloatTensor, 
					columns_to_ignore=None, mv_columns_to_extra_data=None, only_keep_these_columns=None,
					batch_size=32, randomly_shuffle_batch=True, 
					replace_this_entry_with_nan=None, preprocess_data=None):
		'''
			first_row_is_label :  True of False
			row_id_with_feature_names: None, if the label is not the first row 0, set with this
			dataFrame: if dataFrame is set, it ignores the path and use the dataFrame directly as the data itself
			ypath: loads the data as the label
			label_column_name: if the label is loaded together with xpath, this separates label into Y
			preprocess_data: 'center and scale', 'linearly between 0 and 1', 'between 0 and 1 via cdf'
			first_column_as_sample_index: if the first column is used as sample ID
			extra_data: is a list of data that can be appended to the dataset, it has to be the same number of samples as X and y, if not a list, then it will be casted into a list
			extra_data_preprocessing is a list of lists including commands to preprocess, e.g., given 2 extra data [['center and scale'],['ensure_ proper discrete labels']]
			path_prefix: string or list, if list, automatically goes through potential folders where the data might be
			mv_columns_to_extra_data: a list of column names 
		'''
		self.path_prefix = path_prefix
		self.label_column_name = label_column_name
		
		self.label_column_id = label_column_id
		self.randomly_shuffle_batch = randomly_shuffle_batch
		self.row_id_with_feature_names = row_id_with_feature_names
		self.first_row_is_label = first_row_is_label
		self.replace_this_entry_with_nan = replace_this_entry_with_nan
		self.first_column_as_sample_index = first_column_as_sample_index
		self.xpath = xpath
		self.column_names = column_names
		self.only_keep_these_columns = only_keep_these_columns
		self.Y = None
		self.label_type = label_type
		self.batch_size = batch_size
		self.torchloader = None
		self.μ = None
		self.σ = None



		self.mv_columns_to_extra_data = mv_columns_to_extra_data
		self.extra_data_preprocessing = extra_data_preprocessing
		self.extra_data_dictionary = {}
		self.extra_data_dictionary['extra_data'] = extra_data
		self.extra_data_dictionary['numpy'] = self.xDat = []
		self.extra_data_dictionary['df'] = []


		# format the data
		self.format_data_based_on_input_type(X_npArray, dataFrame, columns_to_ignore)
		self.format_extra_data_based_on_input_type()
		self.format_label(Y_npArray=Y_npArray, ypath=ypath, label_column_name=label_column_name, label_column_id=label_column_id, encode_discrete_label_to_one_hot=encode_discrete_label_to_one_hot)
		if self.only_keep_these_columns is not None:
			retained_columns = self.get_columns(self.only_keep_these_columns)
			self.update_data(retained_columns)



		self.columns = self.column_names = self.df.columns		# you can use columns or column_names, they are exactly the same, columns_names is more descriptive, columns is compatible with pandas dataframe
		self.shape = self.df.shape

		self.Data_preprocess(preprocess_data)
		self.initialize_pytorch_settings(xtorchDataType, ytorchDataType)

		self.check_for_missingness()

	def format_extra_data_based_on_input_type(self):
		extra_data = self.extra_data_dictionary['extra_data']
		if extra_data is None and self.mv_columns_to_extra_data is None: return 
		if extra_data is None and self.mv_columns_to_extra_data is not None: extra_data = []								# make sure extra_data is a list 
		if wtype(self.mv_columns_to_extra_data) == 'str': self.mv_columns_to_extra_data = [self.mv_columns_to_extra_data]	# make sure it is in list format


		if wtype(extra_data) == 'str': 		#if string, it is the load path
			pth = wuml.append_prefix_to_path(self.path_prefix, extra_data)
			extra_data = pd.read_csv (pth, header=None)

		if wtype(extra_data) != 'list': 
			extra_data = [extra_data]

		if self.mv_columns_to_extra_data is not None:
			for col in self.mv_columns_to_extra_data:
				extra_data.append(self.pop_column(col))


		if wtype(self.extra_data_preprocessing) == 'str':
			self.extra_data_preprocessing = [[self.extra_data_preprocessing]]


		self.extra_data_dictionary['extra_data'] = extra_data
		for i, data in enumerate(extra_data):
			Dat_np = ensure_numpy(data)
			if self.extra_data_preprocessing is not None:
				for j in self.extra_data_preprocessing[i]:
					if j == 'center and scale':
						Dat_np = preprocessing.scale(Dat_np)
					elif j == 'ensure_ proper discrete labels':
						Dat_np = LabelEncoder().fit_transform(Dat_np)	#Make sure label start from 0

			self.extra_data_dictionary['numpy'].append(ensure_numpy(Dat_np))
			#self.extra_data_dictionary['df'].append(ensure_DataFrame(Dat_np))

	def round(self, rounding=3):
		self.df = self.df.round(decimals=rounding)
		self.update_data(self.df)

	def format_data_based_on_input_type(self, X_npArray, dataFrame, columns_to_ignore):
		first_row_is_label = self.first_row_is_label

		if dataFrame is not None:
			self.df = dataFrame

			if first_row_is_label: 
				self.df = self.df.rename(columns=self.df.iloc[0]).drop(self.df.index[0])
	
			if self.first_column_as_sample_index: 
				self.df = self.df.set_index(list(self.df)[0])

		elif X_npArray is not None:
			if wtype(self.column_names) == 'str': self.column_names = [self.column_names]
			self.df = pd.DataFrame(X_npArray, columns=self.column_names)
			if first_row_is_label: 
				self.df = self.df.rename(columns=self.df.iloc[0]).drop(self.df.index[0])
	
			if self.first_column_as_sample_index: 
				self.df = self.df.set_index(list(self.df)[0])

		else:
			pth = wuml.append_prefix_to_path(self.path_prefix, self.xpath)
			self.df = pd.read_csv (pth, header=None)
			if self.first_column_as_sample_index: self.first_column_as_sample_index = 0
			if first_row_is_label: 
				self.df = pd.read_csv (pth, header=0, index_col=self.first_column_as_sample_index)
			else:
				self.df = pd.read_csv (pth, header=self.row_id_with_feature_names, index_col=self.first_column_as_sample_index)

		if self.replace_this_entry_with_nan is not None:
			self.df = self.df.replace(self.replace_this_entry_with_nan, np.nan)

		self.strip_white_space_from_column_names()
		if columns_to_ignore is not None: self.delete_column(columns_to_ignore)

		self.X = self.df.values


	def format_label(self, Y_npArray=None, ypath=None, label_column_name=None, label_column_id=None, encode_discrete_label_to_one_hot=False):

		# as long as label is needed, we must designate continuous or discrete labels
		if not np.all(np.array([Y_npArray, ypath, label_column_name, label_column_id]) == None):
			if self.label_type is None: raise ValueError('\n\tError : If you are using labels, you must include the argument label_type= "continuout" or "discrete"')

		if Y_npArray is not None:
			self.Y = Y_npArray
		elif ypath is not None: 
			ypth = wuml.append_prefix_to_path(self.path_prefix, ypath)
			self.Y = np.loadtxt(ypth, delimiter=',', dtype=np.float32)			
		elif label_column_name is not None:
			self.Y = self.df[label_column_name].values
			self.delete_column(label_column_name)
		elif label_column_id is not None:
			self.Y = self.df[label_column_id].values
			self.delete_column(label_column_id)

		if self.label_type == 'discrete': 
			self.Y = LabelEncoder().fit_transform(self.Y)	#Make sure label start from 0
			if encode_discrete_label_to_one_hot:
				self.Y = wuml.one_hot_encoding(self.Y)

	def check_for_missingness(self):
		# raise a warning if there are missing entries within the data
		try:
			mL = np.sum(np.isnan(self.X))
			if mL > 0: wuml.jupyter_print('\nWarning: %d entries are missing.'%(mL))
		except: pass

		if self.Y is None: return
		missingLabels = wuml.identify_missing_labels(self.Y)
		if missingLabels is not None: 
			if missingLabels > 0: 
				mL = np.sum(np.isnan(self.Y))
				wuml.jupyter_print('Warning: %.5f percent or %d samples of the labels are missing:  \n'%(missingLabels*100, mL))
				removeSample = input("Would you like to remove the samples with missing labels [y]/n?")
				if removeSample == '' or removeSample == 'y' or removeSample == 'Y':
					wuml.remove_rows_with_missing_labels(self)


	def Data_preprocess(self, preprocess_data='center and scale', mean=None, std=None):
		#	Various ways to preprocess the data
		if preprocess_data == 'center and scale':
			[self.X, self.μ, self.σ] = wuml.center_and_scale(self.X, return_type='ndarray', also_return_mean_and_std=True, mean=mean, std=std)
			self.df = pd.DataFrame(data=self.X, columns=self.df.columns)
		elif preprocess_data == 'linearly between 0 and 1':
			self.X = wuml.wuml.map_data_between_0_and_1(self.X, output_type_name='ndarray', map_type='linear')
			self.df = pd.DataFrame(data=self.X, columns=self.df.columns)
		elif preprocess_data == 'between 0 and 1 via cdf':
			self.X = wuml.wuml.map_data_between_0_and_1(self.X, output_type_name='ndarray', map_type='cdf')
			self.df = pd.DataFrame(data=self.X, columns=self.df.columns)

	def initialize_pytorch_settings(self, xtorchDataType, ytorchDataType):
		#	Property set the data type in Pytorch
		self.xtorchDataType = xtorchDataType				
		if self.label_type == 'discrete': 
			self.ytorchDataType = torch.LongTensor
		else:
			self.ytorchDataType = ytorchDataType				
		self.device = wuml.get_current_device()

	def swap_label(self, column_name_use_for_label):
		labelC = wuml.ensure_DataFrame(self.Y, columns=self.label_column_name)
		newLabel = self.pop_column(column_name_use_for_label)
		self.replace_label(newLabel, label_name=column_name_use_for_label)
		self.append_columns(labelC)

	def replace_label(self, newY, label_name=None):
		self.Y = ensure_numpy(newY)
		self.Y = np.squeeze(self.Y)

		if label_name is not None: self.label_column_name = label_name

	def strip_white_space_from_column_names(self):
		# make sure to strip white space from column names
		update_col_names = []
		for i, col_name in enumerate(self.df.columns):
			if type(self.df.columns[i]).__name__ == 'str':
				update_col_names.append(col_name.strip())
			else:
				update_col_names.append(col_name)

		self.df.columns = update_col_names

	def sort_by(self, column_name, ascending=False):
		self.df = self.df.sort_values(column_name, ascending=ascending)
		self.update_DataFrame(self.df)

	def get_column_names_as_a_list(self):
		return ensure_wData(self.df.columns)

	def get_columns(self, columns):
		if type(columns).__name__ == 'int': 
			remaining_columns = ensure_wData(self.df.iloc[:,columns], column_names=[columns])
		else:	
			columns = ensure_list(columns)
			subColumns = self.df[columns]
			remaining_columns = ensure_wData(subColumns)

		remaining_columns.Y = self.Y
		remaining_columns.extra_data_dictionary = self.extra_data_dictionary
		return remaining_columns

	def update_data(self, data, columns=None):
		if wtype(data) == 'DataFrame':
			self.update_DataFrame(data)
		elif wtype(data) == 'wData':
			self.update_DataFrame(data.df)
		elif wtype(data) == 'ndarray':
			new_df = pd.DataFrame(data, columns=columns)
			self.update_DataFrame(new_df)
		else:
			raise ValueError('wData.update_data function does not recognizes input data type = %s'%wtype(data))


	def update_DataFrame(self, df):
		self.df = df
		self.columns = self.column_names = self.df.columns
		self.X = self.df.values
		self.shape = self.df.shape
	
	def update_column_type(self, update_dictionary):	# the dictionary looks like {'score': float, 'Accᶜ': float } with column name and type
		new_df = self.df.astype(update_dictionary)
		self.update_DataFrame(new_df)


	def reset_index(self):
		self.df.reset_index(drop=True, inplace=True)	
		self.update_DataFrame(self.df)

	def append_rows(self, new_data, reset_index=True):
		df = ensure_DataFrame(new_data, columns=self.columns)
		self.update_DataFrame(pd.concat([self.df,df], axis=0))
		if reset_index: self.reset_index()


	def append_columns(self, new_data, column_names=None):
		df = ensure_DataFrame(new_data, columns=column_names)
		self.update_DataFrame(pd.concat([self.df,df], axis=1))

	def rename_columns(self, column_names):
		if type(column_names).__name__ == 'str':
			column_names = [column_names]

		column_names = np.squeeze(ensure_numpy(column_names)).tolist()
		self.df.columns = column_names
		self.columns = self.column_names = column_names

	def pop_column(self, column_name):
		C = self.get_columns(column_name)
		self.delete_column(column_name)
		return C

	def delete_column(self, column_name):
		if type(column_name) == type([]):
			for name in column_name:
				if name in self.df.columns:
					del self.df[name]

		elif type(column_name) == type(''):
			if column_name in self.df.columns:
				del self.df[column_name]

		elif type(column_name) == type(0):
			if column_name in self.df.columns:
				del self.df[column_name]

		self.columns = self.column_names = self.df.columns
		self.update_DataFrame(self.df)

	def info(self):
		wuml.jupyter_print(self.df.info())

	def get_data_as(self, data_type): #'DataFrame', 'read_csv', 'Tensor'
		if data_type == 'wData': return self
		if data_type == 'Tensor': 
			x = torch.from_numpy(self.df.values)
			x = Variable(x.type(self.xtorchDataType), requires_grad=False)
			X = x.to(self.device, non_blocking=True )
			return X
		if data_type == 'DataFrame': return self.df
		if data_type == 'ndarray': 
			#self.df.values[subset]
			return self.df.values
		if data_type == 'DataLoader':		# and self.torchloader is None 
			self.DM = wuml.DManager(self.df.values, self.Y, self.extra_data_dictionary)
			self.torchloader = DataLoader(dataset=self.DM, batch_size=self.batch_size, shuffle=self.randomly_shuffle_batch, pin_memory=True, num_workers=1)
			return self.torchloader

	def get_all_samples_from_a_class(self, class_name_or_id):
		class_samples = np.empty((0,self.X.shape[1]))
		for i, j in enumerate(self.Y):
			if j == class_name_or_id:
				class_samples = np.vstack((class_samples, self.X[i]))

		class_samples = ensure_wData(class_samples, self.columns)
		N = class_samples.shape[0]
		class_samples.Y = np.full((N), class_name_or_id)
		return class_samples

	def retrieve_scalar_value(self):
		return self.df.to_numpy()[0,0]

	def to_csv(self, path, add_row_indices=False, include_column_names=True, float_format='%.4f'):
		LCn = self.label_column_name

		if LCn is not None:
			self.df[LCn] = self.Y
			self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format)
			self.delete_column(LCn)
		elif self.Y is not None:
			if self.label_column_name is not None:
				self.df[self.label_column_name] = self.Y
				self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format )
				self.delete_column(self.label_column_name)
			else:
				self.df['label'] = self.Y
				self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format )
				self.delete_column('label')
		else:
			self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format)

	def __getitem__(self, item):
		#	If item is string, it will return the column corresponding to the column name
		#	If item is int, it will return the row

		if type(item).__name__ == 'str': 
			return self.get_columns(item)
		elif type(item).__name__ == 'tuple' or type(item).__name__ == 'int' or type(item).__name__ == 'slice': 
			return ensure_wData(self.df.iloc[item])
		else:
			raise ValueError('Error: unrecognized input %s, wData must be [0:3] or [0:2, 0:3] format. You can also get the column by string'%type(item))

	def __str__(self): 
		return str(self.df)

	def __repr__(self): 
		return str(self.df)

	def __iter__(self): 
		self.itr_count = 0
		return self

	def __next__(self):
		''''Returns the next value from team object's lists '''
		try:
			nextItm = ensure_wData(self.df.iloc[self.itr_count].to_frame().transpose())
			#nextItm = self[self.itr_count]
			self.itr_count += 1
			return nextItm
		except:
			self.itr_count = 0
			raise StopIteration


	def plot_2_columns_as_scatter(self, column1, column2):
		if wuml.get_commandLine_input()[1] == 'disabled': return
		X = self.df[column1].to_numpy()
		Y = self.df[column2].to_numpy()
		
		#lp = wuml.scatter(figsize=(10,5))		# (width, height)
		#lp.plot_scatter(X, Y, column1 + ' vs ' + column2, column1, column2, ticker_fontsize=8 )

		lp = wuml.scatter(X, Y, title=str(column1) + ' vs ' + str(column2), 
				xlabel=str(column1), ylabel=str(column2), ticker_fontsize=8)



#!/usr/bin/env python

import os
import sys
if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')

from collections import Counter
from sklearn.datasets import make_classification
from imblearn.over_sampling import SMOTE, RandomOverSampler
import wuml
import numpy as np


class rebalance_data:
	def __init__(self, data, y=None, method='smote'):
		'''method : 'smote', 'oversampling' '''

		self.label_column_name=None
		self.columns = None
		self.label_column_id = None

		if y is None:
			if wuml.wtype(data) == 'wData':
				y = data.Y
				X = data.X
				self.columns = data.columns
				self.label_column_name = data.label_column_name
				self.label_column_id = data.label_column_id
			else:
				raise ValueError('data must be wData if the label y is None')
		else:
			X = wuml.ensure_numpy(data)
			y = wuml.ensure_numpy(y)
			
		if method == 'smote':
			yStats = wuml.get_label_stats(y, print_stat=False)
			sn = yStats.get_columns('num sample')
			smallest_class_sample_num =  np.min(sn.X)
			if smallest_class_sample_num < 6: raise ValueError('Error: rebalance data with smote must have at least 6 samples in the smallest class.')
	
			sm = SMOTE(random_state=42)
			X_res, y_res = sm.fit_resample(X, y)
		elif method == 'oversampling':
			ROS = RandomOverSampler()
			X_res, y_res = ROS.fit_resample(X, y)


		self.balanced_data = wuml.wData(X_npArray=X_res, Y_npArray=y_res, first_row_is_label=False, label_type='discrete',
										column_names=self.columns, label_column_name=self.label_column_name, label_column_id=self.label_column_id)


if __name__ == "__main__":
	X1 = np.random.randn(20, 2) + 5
	X2 = np.random.randn(7, 2) - 5
	X = np.vstack((X1,X2))
	y = np.append(np.ones(20), np.zeros(7))

	rebalancer = rebalance_data(X, y, method='oversampling')
	print(X,'\n')

	print(rebalancer.balanced_data.shape)
	print(rebalancer.balanced_data)
	import pdb; pdb.set_trace()



import numpy as np
from wuml.type_check import *


def norm(data, norm='l2'):
	# If norm='l1':
	#	If data is a vector: compute l1 norm
	#	If data is a matrix: compute l1 norm for each row
	# If norm='l2':
	#	If data is a vector: compute l2 norm
	#	If data is a matrix: compute l2 norm for each row
	# If norm='fro':
	#	If data is a vector: compute frobenius norm
	#	If data is a matrix: compute frobenius norm 
	# If norm='nuc':
	#	If data is a vector: compute nuclear norm
	#	If data is a matrix: compute nuclear norm 

	X = np.squeeze(ensure_numpy(data))

	if norm=='l1':
		if len(X.shape) == 1:
			return np.linalg.norm(X, ord=1)
		else:
			return np.linalg.norm(X, ord=1, axis=1)
	elif norm=='l2':
		if len(X.shape) == 1:
			return np.linalg.norm(X, ord=2)
		else:
			return np.linalg.norm(X, ord=2, axis=1)
	elif norm=='fro':
		return np.linalg.norm(X, ord='fro')
	elif norm=='nuc' or norm=='nuclear':
		return np.linalg.norm(X, ord='nuc')





import numpy as np
import wuml
import torch
import pandas as pd
from torch.autograd import Variable


def wtype(data):
	return type(data).__name__

def cast_each_item_in_list_as(list_of_data, cast_type):
	casted_list = [] 
	for dat in list_of_data:
		casted_dat = ensure_data_type(dat, type_name=cast_type)
		casted_list.append(casted_dat)

	return casted_list

def get_function_name(function):
	if callable(function):
		return function.__name__
	else:
		return 'None'

def type_check_with_error(data, desired_type, function_name=''):
	if wtype(data) != desired_type:
		raise ValueError('For function %s, you must input type as %s'%(function_name, desired_type))


def ensure_column(data):
	X = ensure_numpy(data)
	if len(X.shape) == 1:
		X = X.expand_dim().T

	elif len(X.shape) == 2:
		if X.shape[0] == 1:
			X = X.T
	return X

def is_binary_label(data):
	y = ensure_numpy(data)
	y_labels = np.unique(y)

	if len(y_labels) == 2 and (1 in y_labels) and (0 in y_labels):
		return True
	else:
		return False

def ensure_label(data, y=None, y_column_name=None):
	NP = ensure_numpy
	S = np.squeeze

	if y is not None:
		Y = S(NP(y))
	elif y_column_name is not None:
		Y = data[y_column_name].values
	elif wtype(data) == 'wData':
		Y = data.Y
	else: raise ValueError('Undefined label Y')

	return Y

def ensure_data_type(data, type_name='ndarray', ensure_column_format=True, column_names=None):

	if type_name=='ndarray':
		return ensure_numpy(data, ensure_column_format=ensure_column_format)
	elif type_name=='DataFrame':
		return ensure_DataFrame(data, columns=column_names)
	elif type_name=='Tensor':
		return ensure_tensor(data)
	elif type_name=='wData':
		return ensure_wData(data, column_names=column_names)


def ensure_list(data):
	if wtype(data) == 'str': 
		return [data]
	elif wtype(data) == 'list': 
		return data
	elif wtype(data) == 'NoneType': 
		return data
	elif wtype(data) == 'Int64Index': 
		return data.values.tolist()
	elif wtype(data) == 'Index': 	
		return data.tolist()
	elif wtype(data) == 'ndarray': 
		return data.tolist()
	elif wtype(data) == 'wData': 
		npD = data.df.to_numpy()
		npD = np.squeeze(npD)
		return npD.tolist()
	elif type_name=='DataFrame':
		npD = data.to_numpy()
		npD = np.squeeze(npD)
		return npD.tolist()


def ensure_wData(data, column_names=None, extra_data=None):
	if wtype(data) == 'ndarray': 
		return wuml.wData(X_npArray=data, column_names=column_names, extra_data=extra_data)
	elif wtype(data) == 'Index': 
		ArrayType = np.array(data.tolist())
		return wuml.wData(X_npArray=ArrayType, column_names=column_names, extra_data=extra_data)
	elif wtype(data) == 'wData': 
		return data
	elif wtype(data) == 'DataFrame': 
		return wuml.wData(dataFrame=data, column_names=column_names, extra_data=extra_data)
	elif wtype(data) == 'Tensor': 
		X = data.detach().cpu().numpy()
		return wuml.wData(X_npArray=X, column_names=column_names, extra_data=extra_data)
	elif wtype(data) == 'dimension_reduction': 
		return wuml.wData(X_npArray=data.Ӽ, extra_data=extra_data)
	elif wtype(data) == 'Series': 
		return wuml.wData(X_npArray=data.to_numpy(), column_names=column_names, extra_data=extra_data)
	elif wtype(data) == 'float64': 
		return wuml.wData(X_npArray=np.array([data]), column_names=column_names, extra_data=extra_data)


def ensure_DataFrame(data, columns=None, index=None):
	columns = ensure_list(columns)

	if wtype(data) == 'ndarray': 
		df = pd.DataFrame(data, columns=columns, index=index)
	elif wtype(data) == 'wData': 
		df = data.df
		df.columns = data.df.columns
		df.index = data.df.index
	elif wtype(data) == 'DataFrame': 
		df = data
		if columns is not None: df.columns = columns
		#df.index = data.df.index
	elif wtype(data) == 'Tensor': 
		X = data.detach().cpu().numpy()
		df = pd.DataFrame(X)
		df.columns = data.df.columns
		df.index = data.df.index
	elif wtype(data) == 'dimension_reduction': 
		df = pd.DataFrame(data.Ӽ)

	return df


def ensure_numpy(data, rounding=None, ensure_column_format=True):

	if wtype(data) == 'ndarray': 
		if ensure_column_format and len(data.shape) == 1:
			X = np.atleast_2d(data).T
		else:
			X = data
	elif wtype(data) == 'Index': 
		return data.to_numpy()
	elif wtype(data) == 'list': 
		return np.array(data)
	elif wtype(data) == 'wData': 
		X = data.df.values
	elif wtype(data) == 'DataFrame': 
		X = data.values
	elif wtype(data) == 'Tensor': 
		X = data.detach().cpu().numpy()
	elif wtype(data) == 'Series': 
		X = data.values
	elif wtype(data) == 'Int64Index': 
		return data.to_numpy()
	elif np.isscalar(data):
		X = np.array([[data]])
	elif wtype(data) == 'dimension_reduction': 
		X = data.Ӽ
	else:
		raise ValueError('Unknown dataType %s'%wtype(data))

	if rounding is not None: X = np.round(X, rounding)
	return X


def ensure_tensor(data, dataType=torch.FloatTensor):
	device = wuml.get_current_device()

	if wtype(data) == 'ndarray': 
		x = torch.from_numpy(data)
		x = Variable(x.type(dataType), requires_grad=False)
		X = x.to(device, non_blocking=True )
	elif wtype(data) == 'wData': 
		X = data.get_data_as('Tensor')
	elif wtype(data) == 'DataFrame': 
		X = data.values
	elif np.isscalar(data):
		X = np.array([[data]])
	elif wtype(data) == 'Tensor': 
		x = Variable(data.type(dataType), requires_grad=False)
		X = x.to(device, non_blocking=True )
	elif wtype(data) == 'dimension_reduction': 
		x = torch.from_numpy(data.Ӽ)
		x = Variable(x.type(dataType), requires_grad=False)
		X = x.to(device, non_blocking=True )

	else:
		raise ValueError('Unknown dataType %s'%wtype(data))

	return X


def ensure_proper_model_input_format(data):
	X = ensure_numpy(data)	
	if len(X.shape) == 2:
		if X.shape[1] == 1:
			X = X.T
	if len(X.shape) == 1:
		X = np.expand_dims(X, axis=0)

	return X





if __name__ == "__main__":

	y = [1, 0, 0, 0, 1]
	print(is_binary_label(y))

	y = [True, False, False, True, True]
	print(is_binary_label(y))

	y = ['red', 'red', 'red', 'blue', 'blue']
	print(is_binary_label(y))




from wuml.basicNetwork import *
from wuml.type_check import *
import torch
import numpy as np

#	x -> encoder -> x̂
#	x̂ -> midcoder -> ẙ	
#	x̂ -> decoder -> ŷ	
#	possible autoencoder objective λ could be 0
#	loss = (x - ŷ)ᒾ + λ * objective(ẙ, y)

class autoencoder():
	def __init__(self, bottleneck_size, X, Y=None, mask=None, costFunction=None,
						EncoderStructure=None, DecoderStructure=None, encoder_output_weight_structure=None, 
						default_depth=4, default_activation_function='relu',
						early_exit_loss_threshold=0.0000001, pickled_network_info=None,
						on_new_epoch_call_back = None, max_epoch=1000, 	
						X_dataType=None, Y_dataType=None, network_usage_output_type='Tensor',
						learning_rate=0.001):
		'''
			X : This should be wData type
			possible activation functions: softmax, relu, tanh, sigmoid, none
			simplify_network_for_storage: if a network is passed as this argument, we create a new network strip of unnecessary stuff
		'''
		if get_commandLine_input()[1] == 'disabled': max_epoch = 10
		self.on_new_epoch_call_back = on_new_epoch_call_back #set this as a callback at each function

		# 	if we are loading a saved network from pickle, the behavior would be different
		if pickled_network_info is None:
			self.X = X
			bottleneck_size = int(bottleneck_size)
	
			if EncoderStructure is None or DecoderStructure is None:
				[self.encoder_structure, self.decoder_structure] = self.get_default_encoder_decoder(X, bottleneck_size, default_depth, default_activation_function)
			else:
				self.encoder_structure = EncoderStructure
				self.decoder_structure = DecoderStructure
	
	
			self.encoder = flexable_Model(X.shape[1], self.encoder_structure)
			self.decoder = flexable_Model(bottleneck_size, self.decoder_structure)
			if encoder_output_weight_structure is None:
				self.midcoder = None
			else:
				self.midcoder = flexable_Model(bottleneck_size, encoder_output_weight_structure)
	
			self.trainLoader = X.get_data_as('DataLoader')
	
			self.network_usage_output_type = network_usage_output_type
			self.lr = learning_rate
			self.max_epoch = max_epoch
			self.early_exit_loss_threshold = early_exit_loss_threshold
	
			# use the default wData type unless strictly set to something else
			# if label is discrete, Y should default set to torch.LongTensor for integer values
			if X_dataType is not None: self.X_dataType = X_dataType
			else: self.X_dataType = X.xtorchDataType
			if Y_dataType is not None: self.Y_dataType = Y_dataType
			else: self.Y_dataType = X.ytorchDataType
	
	
			if costFunction is None: costFunction = torch.nn.MSELoss()
			else: self.costFunction = costFunction
			self.network_output_in_CPU_during_usage = False
		else:
			pDat = pickled_network_info
			self.costFunction = pDat['costFunction']
	
			self.lr = pDat['lr']
			self.max_epoch = pDat['max_epoch']
			self.encoder_structure = pDat['encoder_structure']
			self.decoder_structure = pDat['decoder_structure']
	
			self.encoder = pDat['encoder']
			self.decoder = pDat['decoder']
			self.midcoder = pDat['midcoder']

		self.device = wuml.get_current_device()
		self.encoder.to(self.device)		# store the network weights in gpu or cpu device
		self.decoder.to(self.device)		# store the network weights in gpu or cpu device
		if self.midcoder is not None:
			self.midcoder.to(self.device)

		self.info()


	def output_network_data_for_storage(self):
		net = {}
		net['name'] = self.__class__.__name__
		net['costFunction'] = marshal.dumps(self.costFunction.__code__)

		net['lr'] = self.lr
		net['max_epoch'] = self.max_epoch
		net['X_dataType'] = self.X_dataType
		net['Y_dataType'] = self.Y_dataType

		net['encoder_structure'] = self.encoder_structure
		net['decoder_structure'] = self.decoder_structure

		net['encoder'] = self.encoder
		net['decoder'] = self.decoder
		net['midcoder'] = self.midcoder

		return net


	def info(self, printOut=True):
		 
		info_str ='Autoencoder Info:\n'
		info_str += '\tLearning rate: %.3f\n'%self.lr
		info_str += '\tMax number of epochs: %d\n'%self.max_epoch
		info_str += '\tCost Function: %s\n'%wuml.get_function_name(self.costFunction)
		if self.on_new_epoch_call_back is not None: info_str += '\tTrain Loop Callback: %s\n'%wuml.get_function_name(self.on_new_epoch_call_back)
		info_str += '\tDevice type: %r\n'%self.device
		info_str += '\tEncoder Structure\n'
		for i in self.encoder.children():
			try: info_str += ('\t\t%s , %s\n'%(i,i.activation))
			except: info_str += ('\t\t%s \n'%(i))

		if self.midcoder is not None:
			info_str += '\tEncoder Extra Output weight Structure\n'
			for i in self.midcoder.children():
				try: info_str += ('\t\t%s , %s\n'%(i,i.activation))
				except: info_str += ('\t\t%s \n'%(i))


		info_str += '\tEncoder Structure\n'
		for i in self.decoder.children():
			try: info_str += ('\t\t%s , %s\n'%(i,i.activation))
			except: info_str += ('\t\t%s \n'%(i))

		if printOut: wuml.jupyter_print(info_str)
		return info_str


	def get_default_encoder_decoder(self, X, bottleneck_size, default_depth, default_activation_function):
		# the default encoder and decoder depth are 3 layers
		d = X.X.shape[1]

		#step =  int(np.absolute(d - bottleneck_size)/default_depth)
		sizes = np.floor(np.linspace(d, bottleneck_size, default_depth))
		Rsizes = np.flip(sizes)

		sizes = np.delete(sizes, 0)	# remove the input dimension since it is already know
		Rsizes = np.delete(Rsizes, 0)

		encoder_structure = []
		decoder_structure = []

		for i in sizes:
			if i == bottleneck_size: encoder_structure.append((int(i),'none'))
			else: 
				encoder_structure.append((int(i),default_activation_function))
				#encoder_structure.append(('bn', True))

		for i in Rsizes:
			if i == d: decoder_structure.append((int(i),'none'))
			else: 
				decoder_structure.append((int(i),default_activation_function))
				#decoder_structure.append(('bn', True))

		return [encoder_structure, decoder_structure]


	def fit(self, print_status=True):
		enc_param = self.encoder.parameters()
		dec_param = self.decoder.parameters()

		[ℓ, TL, mE, Dev] = [self.costFunction, self.trainLoader, self.max_epoch, self.device]
		[Xtype, Ytype] = [self.X_dataType, self.Y_dataType]

		self.run_autoencoder_SGD(ℓ, enc_param, dec_param, TL, Dev, encoder=self.encoder, decoder=self.decoder, lr=self.lr, print_status=print_status,
				max_epoch=mE, X_dataType=Xtype, Y_dataType=Ytype, early_exit_loss_threshold=self.early_exit_loss_threshold,
				on_new_epoch_call_back = self.on_new_epoch_call_back)


	def objective_network(self, data, output_type=None):
		inputType = wtype(data)
		
		x = ensure_proper_model_input_format(data)
		x = ensure_tensor(x, dataType=torch.FloatTensor)

		x̂ = self.encoder(x)
		ẙ = self.midcoder(x̂)
		ẙ = ensure_data_type(ẙ, type_name=inputType)
		if wtype(ẙ) == 'wData': ẙ.Y = data.Y

		return ẙ

	def reduce_dimension(self, data, output_type='wData', outPath=None):
		x = ensure_tensor(data, dataType=torch.FloatTensor)
		x̂ = self.encoder(x)

		x̂ = ensure_data_type(x̂, type_name=output_type)
		#if output_type == 'ndarray' or self.network_usage_output_type == 'ndarray':
		#	x̂ = x̂.detach().cpu().numpy()
		#elif output_type == 'wData':
		#	x̂ = ensure_wData(x̂)
		#elif self.network_output_in_CPU_during_usage:
		#	x̂ = x̂.detach().cpu()

		if outPath is not None:
			wD = ensure_wData(x̂)
			wD.to_csv(outPath)

		if wtype(x̂) == 'wData': x̂.Y = data.Y
		return x̂

	def __call__(self, data, output_type='Tensor', out_structural=None):
		'''
			out_structural (mostly for classification purpose): None, '1d_labels', 'one_hot'
		'''

		x = ensure_tensor(data, dataType=torch.FloatTensor)
		x̂ = self.encoder(x)
		ŷ = self.decoder(x̂)

		if output_type == 'ndarray' or self.network_usage_output_type == 'ndarray':
			return ŷ.detach().cpu().numpy()
		elif self.network_output_in_CPU_during_usage:
			return ŷ.detach().cpu()

		return ŷ



	def run_autoencoder_SGD(self, loss_function, enc_param, dec_param, trainLoader, device, early_exit_loss_threshold=0.000000001,
					X_dataType=torch.FloatTensor, Y_dataType=torch.FloatTensor, 
					encoder=None, decoder=None, lr=0.001, print_status=True, max_epoch=1000,
					on_new_epoch_call_back=None):
	
	
		enc_optimizer = torch.optim.Adam(enc_param, lr=lr)	
		dec_optimizer = torch.optim.Adam(dec_param, lr=lr)	

		enc_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau( enc_optimizer, factor=0.5, min_lr=1e-10, patience=50, verbose=False)
		dec_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau( dec_optimizer, factor=0.5, min_lr=1e-10, patience=50, verbose=False)
	
		#	if mid coder is used: midcoder takes output of bottleneck for separate process
		if self.midcoder is None:
			midcoder_param = None
			mid_optimizer = None
			mid_scheduler = None
		else:
			midcoder_param = self.midcoder.parameters()
			mid_optimizer = torch.optim.Adam(midcoder_param, lr=lr)	
			mid_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau( dec_optimizer, factor=0.5, min_lr=1e-10, patience=50, verbose=False)

		for epoch in range(max_epoch):
	
			loss_list = []	
			recons_loss_list = []	
			objective_loss_list = []	
			for (i, data) in enumerate(trainLoader):
				[x, y, ind] = data
	
				x = Variable(x.type(X_dataType), requires_grad=False)
				y = Variable(y.type(Y_dataType), requires_grad=False)
				x= x.to(device, non_blocking=True )
				y= y.to(device, non_blocking=True )
				enc_optimizer.zero_grad()
	
				x̂ = encoder(x)
				ŷ = decoder(x̂)

				if self.midcoder is not None:
					ẙ = self.midcoder(x̂)
					ẙ = torch.squeeze(ẙ)
				else: ẙ = 0

				ŷ = torch.squeeze(ŷ)
				y = torch.squeeze(y)

				all_losses = loss_function(x, x̂, ẙ, y, ŷ, ind)
				if wtype(all_losses) == 'list': 
					if len(all_losses) == 3:
						[loss, Reconst_loss, obj_loss] = all_losses

					recons_loss_list.append(Reconst_loss.item())
					objective_loss_list.append(obj_loss.item())
				else: loss = all_losses

				if torch.isnan(loss): import pdb; pdb.set_trace()
	
				loss.backward()
				enc_optimizer.step()
				dec_optimizer.step()
				if mid_optimizer is not None: mid_optimizer.step()

				loss_list.append(loss.item())
	
			loss_avg = np.array(loss_list).mean()
			enc_scheduler.step(loss_avg)
			dec_scheduler.step(loss_avg)	
			if mid_scheduler is not None: mid_scheduler.step(loss_avg)
			
			if loss_avg < early_exit_loss_threshold: 
				print('\n')
				break;
			if print_status:
				if get_commandLine_input()[1] != 'disabled': 
					if wtype(all_losses) == 'list':
						l1 = loss_avg
						l2 = np.array(recons_loss_list).mean()
						l3 = np.array(objective_loss_list).mean()
	
						txt = '\tepoch: %d, Total Loss/dimension: %.4f, Reconstruct loss: %.4f, Objective loss: %.4f, Learning Rate: %.8f'%((epoch+1), l1, l2, l3, enc_scheduler._last_lr[0])
						write_to_current_line(txt)
					else:
						txt = '\tepoch: %d, Avg Loss/dimension: %.4f, Learning Rate: %.8f'%((epoch+1), loss_avg, enc_scheduler._last_lr[0])
						write_to_current_line(txt)
	
			if on_new_epoch_call_back is not None:
				on_new_epoch_call_back(loss_avg, (epoch+1), enc_scheduler._last_lr[0])
	
			#if model is not None:
			#	if "on_new_epoch" in dir(model):
			#		early_exit = model.on_new_epoch(loss_avg, (epoch+1), scheduler._last_lr[0])
			#		if early_exit: break


